from ffxivcalc.Jobs.Base_Spell import DOTSpell, ManaRequirement, buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Healer.Healer_Spell import AstrologianSpell
from copy import deepcopy
import logging
main_logging = logging.getLogger("ffxivcalc")
ast_logging = main_logging.getChild("astrologian")
Lock = 0

#Requirement

def DrawRequirement(Player, Spell):
    return Player.DrawStack > 0, Player.DrawCD

def RedrawRequirement(Player, Spell):
    return Player.Redraw, -1

def ArcanumRequirement(Player, Spell):
    return Player.HasCard, -1

def MinorArcanaRequirement(Player, Spell):
    return Player.MinorArcanaCD <= 0, Player.MinorArcanaCD

def LordOfCrownRequirement(Player, Spell):
    return Player.LordOfCrown, -1 #Will be assumed to be given by Minor Arcana

def DivinationRequirement(Player, Spell):
    return Player.DivinationCD <= 0, Player.DivinationCD

def LightspeedRequirement(Player, Spell):
    return Player.LightspeedCD <= 0, Player.LightspeedCD

def MacrocosmosRequirement(Player, Spell):
    return Player.MacrocosmosCD <= 0, Player.MacrocosmosCD

def MicrocosmosRequirement(Player, Spell):
    return Player.MacrocosmosCD > 165, -1 #Since Micro is only available after 15 sec

def ExaltationRequirement(Player, Spell):
    return Player.ExaltationCD <= 0, Player.ExaltationCD

def NeutralSectRequirement(Player, Spell):
    return Player.NeutralSectCD <= 0, Player.NeutralSectCD

def HoroscopeRequirement(Player, Spell):
    return Player.HoroscopeCD <= 0, Player.HoroscopeCD

def CelestialIntersectionRequirement(Player, Spell):
    return Player.CelestialIntersectionStack > 0, Player.CelestialIntersectionCD

def EarthlyStarRequirement(Player, Spell):
    return Player.EarthlyStarCD <= 0, Player.EarthlyStarCD

def StellarDetonationRequirement(Player, Spell):
    if Player.EarthlyStarCD > 50 : Spell.Potency = 205 #Changes potency with respect to how long the player has waited since doing EarthlyStar
    elif Player.EarthlyStarCD <= 50 and Player.EarthlyStarCD >= 40 : Spell.Potency = 310
    return Player.EarthlyStarCD > 40, -1

def CelestialOppositionRequirement(Player, Spell):
    return Player.CelestialOppositionCD <= 0, Player.CelestialOppositionCD

def CollectiveUnconsciousRequirement(Player, Spell):
    return Player.CollectiveUnconsciousCD <= 0, Player.CollectiveUnconsciousCD

def SynastryRequirement(Player, Spell):
    return Player.SynastryCD <= 0, Player.SynastryCD

def EssentialDignityRequirement(Player, Spell):
    return Player.EssentialDignityStack >0, Player.EssentialDignityCD

#Apply

def ApplyEssentialDignity(Player, Enemy):
    if Player.EssentialDignityStack == 2:
        Player.EffectCDList.append(EssentialDignityStackCheck)
        Player.EssentialDignityCD = 40
    Player.EssentialDignityStack -= 1

def ApplySynastry(Player, Enemy):
    Player.SynastryCD = 120

def ApplyCollectiveUnconscious(Player, Enemy):
    Player.CollectiveUnconsciousCD = 60

def ApplyCelestialOpposition(Player, Enemy):
    Player.CelestialOppositionCD = 60

def ApplyEarthlyStar(Player, Enemy):
    Player.EarthlyStarCD = 60

def ApplyCelestialIntersection(Player, Enemy):

    if Player.CelestialIntersectionStack == 2:
        Player.EffectCDList.append(CelestialIntersectionStackCheck)
        Player.CelestialIntersectionCD = 30
    
    Player.CelestialIntersectionStack -= 1


def ApplyHoroscope(Player, Enemy):
    Player.HoroscopeCD = 60

def ApplyNeutralSect(Player, Enemy):
    Player.NeutralSectCD = 120

def ApplyExaltation(Player, Enemy):
    Player.ExaltationCD = 60

def ApplyMacrocosmos(Player, Enemy):
    Player.MacrocosmosCD = 180

def ApplyAstrodyne(Player, Enemy):
    check = 0

    if Player.Lunar:
        check +=1
        Player.Lunar = False
    if Player.Solar:
        check +=1
        Player.Solar = False
    if Player.Celestial:
        check +=1
        Player.Celestial = False
    #Will only consider Body and Mind
    if check >= 2:
        Player.EffectList.append(BodyEffect)
        Player.Haste += 10
        Player.hasteHasChanged = True
        Player.hasteChangeValue = 10
        Player.EffectCDList.append(BodyCheck)#Only 1 check for both Body and Mind since same Timer
        Player.BodyTimer = 15
    if check == 3:
        Player.buffList.append(AstrodyneBuff)
        Player.EffectCDList.append(MindCheck)

                                        # Only relevant to PreBakedAction and only does that code if true
        #if Player.CurrentFight.SavePreBakedAction:
        #    fight = Player.CurrentFight
        #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15, AstrodyneBuff.MultDPS)
        #    Player.PercentBuffHistory.append(history)

def ApplyDraw(Player, Enemy):
    Player.HasCard = True
    Player.Redraw = True

    if Player.DrawStack == 2:
        Player.EffectCDList.append(DrawStackCheck)
        Player.DrawCD = 30
    Player.DrawStack -= 1

def ApplyRedraw(Player, Enemy):
    Player.Redraw = False

def ApplyMinorArcana(Player, Enemy):
    Player.LordOfCrown = True
    Player.MinorArcanaCD = 60

def ApplyLordOfCrown(Player, Enemy):
    Player.LordOfCrown = False

def ApplyDivination(Player, Enemy):
    Player.DivinationCD = 120
    Player.DivinationTimer = 15
    Player.EffectCDList.append(DivinationCheck)
    Enemy.buffList.append(DivinationBuff) #Just give DPS bonus on Enemy instead of raid wide buff


                             # Only relevant to PreBakedAction and only does that code if true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15, DivinationBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyLightspeed(Player, Enemy):
    Player.LightspeedTimer = 15
    Player.LightspeedCD = 90
    Player.EffectList.append(LightspeedEffect)
    Player.EffectCDList.append(LightspeedCheck)

def ApplyCombust(Player, Enemy):
    if Player.CumbustDOT == None:
        Player.CumbustDOT = deepcopy(CumbustDOT)
        Player.CumbustDOTTimer = 30
        Player.DOTList.append(Player.CumbustDOT)
        Player.EffectCDList.append(CumbustDOTCheck)
    else : Player.CumbustDOT.resetBuffSnapshot() # If already applied reset snapshot
    Player.CumbustDOTTimer = 30

#Effect

def BodyEffect(Player, Spell):
    pass

def LightspeedEffect(Player, Spell):
    if Spell.GCD: Spell.CastTime = max(Lock, Spell.CastTime - 2.5)

#Check

def EssentialDignityStackCheck(Player, Enemy):
    if Player.EssentialDignityCD <= 0:
        if Player.EssentialDignityStack == 1:
            Player.EffectToRemove.append(EssentialDignityStackCheck)
        else:
            Player.EssentialDignityCD = 40
        Player.EssentialDignityStack += 1

def BodyCheck(Player, Enemy):
    if Player.BodyTimer <= 0:
        Player.Haste -= 10
        Player.hasteHasChanged = True
        Player.hasteChangeValue = -10
        Player.EffectList.remove(BodyEffect)
        Player.EffectToRemove.append(BodyCheck)

def MindCheck(Player, Enemy):
    if Player.BodyTimer <= 0:
        Player.buffList.remove(AstrodyneBuff)
        Player.EffectToRemove.append(MindCheck)

def DrawStackCheck(Player, Enemy):
    if Player.DrawCD <= 0:
        if Player.DrawStack == 1:
            Player.EffectToRemove.append(DrawStackCheck)
        else:
            Player.DrawCD = 30
        Player.DrawStack += 1

def DivinationCheck(Player, Enemy):
    if Player.DivinationTimer <= 0:
        Enemy.buffList.remove(DivinationBuff)
        Player.EffectToRemove.append(DivinationCheck)

def LightspeedCheck(Player, Enemy):
    if Player.LightspeedTimer <= 0:
        Player.EffectList.remove(LightspeedEffect)
        Player.EffectToRemove.append(LightspeedCheck)


def CumbustDOTCheck(Player, Enemy):
    if Player.CumbustDOTTimer <= 0:
        Player.DOTList.remove(Player.CumbustDOT)
        Player.CumbustDOT = None
        Player.EffectToRemove.append(CumbustDOTCheck)

def CelestialIntersectionStackCheck(Player, Enemy):
    if Player.CelestialIntersectionCD <= 0:
        if Player.CelestialIntersectionStack == 1:
            Player.EffectToRemove.append(CelestialIntersectionStackCheck)
        else:
            Player.CelestialIntersectionCD = 30
        Player.CelestialIntersectionStack += 1

#GCD
Malefic = AstrologianSpell(25871, True, 1.5, 2.5, 250, 400, empty, [ManaRequirement], type = 1)
Combust = AstrologianSpell(16554, True, Lock, 2.5, 0, 400, ApplyCombust, [ManaRequirement], type = 1)
LordOfCrown = AstrologianSpell(7444, True, Lock, 1, 250, 0,  ApplyLordOfCrown, [LordOfCrownRequirement], type = 1)
CumbustDOT = DOTSpell(-12, 55, False)
Gravity = AstrologianSpell(25872, True, 1.5, 2.5, 130, 400, empty, [ManaRequirement], type = 1)
#Heal GCD
AspectedHelios = AstrologianSpell(3601, True, 1.5, 2.5, 0, 800, empty, [ManaRequirement], type = 1, AOEHeal=True)
AspectedBenefic = AstrologianSpell(3595, True, 0, 2.5, 0, 400, empty, [ManaRequirement], type = 1, TargetHeal=True)
Benefic2 = AstrologianSpell(3610, True, 1.5, 2.5, 0, 700, empty, [ManaRequirement], type = 1, TargetHeal=True)
Benefic = AstrologianSpell(3594, True, 1.5, 2.5, 0, 400, empty, [ManaRequirement], type = 1, TargetHeal=True)
Helios = AstrologianSpell(3600, True, 1.5, 2.5, 0, 700, empty, [ManaRequirement], type = 1, AOEHeal=True)
EssentialDignity = AstrologianSpell(3614, False, 0, 0, 0, 0, ApplyEssentialDignity, [EssentialDignityRequirement], type = 1, TargetHeal=True)
Macrocosmos = AstrologianSpell(25874, True, 0, 0, 250, 600, ApplyMacrocosmos, [MacrocosmosRequirement, ManaRequirement], type = 1)
Microcosmos = AstrologianSpell(17, True, 0, 0, 0, 0, empty, [MicrocosmosRequirement], type = 1, AOEHeal=True)
LadyOfCrown = AstrologianSpell(7445, True, 0, 1, 0, 0, empty, [], type = 1)
Ascend = AstrologianSpell(3603, True, 8, 2.5, 0, 2400, empty, [ManaRequirement])
#oGCD
Lightspeed = AstrologianSpell(3606, False, Lock, 0, 0, 0, ApplyLightspeed, [LightspeedRequirement])
Divination = AstrologianSpell(16552, False, Lock, 0, 0, 0, ApplyDivination, [DivinationRequirement])
MinorArcana = AstrologianSpell(7443, False, Lock, 0, 0, 0, ApplyMinorArcana, [MinorArcanaRequirement])
Draw = AstrologianSpell(3590, False, Lock,0, 0, 0, ApplyDraw, [DrawRequirement])
Redraw = AstrologianSpell(3593, False, Lock, 0, 0, 0, ApplyRedraw, [RedrawRequirement])
Astrodyne = AstrologianSpell(25870, False, Lock, 0, 0, 0, ApplyAstrodyne, [])
#Heal oGCD
Exaltation = AstrologianSpell(25873, False, 0, 0, 0, 0, ApplyExaltation, [ExaltationRequirement])
NeutralSect = AstrologianSpell(16559, False, 0, 0, 0, 0, ApplyNeutralSect, [NeutralSectRequirement])
Horoscope = AstrologianSpell(16558, False, 0, 0, 0, 0, ApplyHoroscope, [HoroscopeRequirement])
CelestialIntersection = AstrologianSpell(16556, False, 0, 0, 0, 0, ApplyCelestialIntersection, [CelestialIntersectionRequirement], TargetHeal=True)
EarthlyStar = AstrologianSpell(7439, False, 0, 0, 0, 0, ApplyEarthlyStar, [EarthlyStarRequirement])
StellarDetonation = AstrologianSpell(8324, False, 0, 0, 0, 0, empty, [StellarDetonationRequirement])
CelestialOpposition = AstrologianSpell(16553, False, 0, 0, 0, 0, ApplyCelestialOpposition, [CelestialOppositionRequirement], AOEHeal=True)
CollectiveUnconscious = AstrologianSpell(3613, False, 0, 0, 0, 0, ApplyCollectiveUnconscious, [CollectiveUnconsciousRequirement])
Synastry = AstrologianSpell(3612, False, 0, 0, 0, 0, ApplySynastry, [SynastryRequirement])

#Arcanum require a target within the team, so it will be a function that will return a spell that
#will target the given player. It is also assumed that the bonus is 6%
ArcanumBuffMax = buff(1.06,name="Arc")
ArcanumBuffMin = buff(1.03,name="Arc")
AstrodyneBuff = buff(1.05,name="Astro")
DivinationBuff = buff(1.06,name="Div")


def Arcanum(Target, Type, Melee):
    """
    This function generates an AstrologianSpell corresponding to a card being played.
    Target : Player -> Reference of the player object being targetted
    Type : string -> Type of the card played
    Melee : bool -> True if this card gives 6% to melees and false otherwise (if false then it gives 6% to ranged)
    """

    def ArcanumCheck(Player, Enemy):
        if Player.ArcanumTimer <= 0:
            Player.buffList.remove(Player.ArcanumBuff)
            Player.ArcanumBuff = None
            Player.EffectToRemove.append(ArcanumCheck)

    def ApplyArcanum(Player, Spell):
                             # Refreshing targetID
        ArcanumSpell.TargetID = Target.playerID
        Player.HasCard = False
                             # Default value for buff
        buff = ArcanumBuffMin

                             # Check if the target is a melee or ranged and has the 6% buff
        if (Melee and (Target.RoleEnum == 3 or Target.RoleEnum == 4)):
            ast_logging.debug("Giving to melee")
            buff = ArcanumBuffMax
        elif (not Melee and (Target.RoleEnum == 1 or Target.RoleEnum == 2 or Target.RoleEnum == 5)):
            ast_logging.debug("giving to ranged")
            buff = ArcanumBuffMax
        else:
            ast_logging.debug("Not given to optimal job")
                             # The target already has a card given to him
        if Target.ArcanumBuff != None:
            Target.buffList.remove(Target.ArcanumBuff)
        else:                # If it already didn't have a card. We must give the target the Check
            Target.EffectCDList.append(ArcanumCheck)

                             # Given card buff to the target
        Target.ArcanumBuff = deepcopy(buff)
        Target.buffList.append(Target.ArcanumBuff)
        Target.ArcanumTimer = 15

        if Type == "Lunar" : Player.Lunar = True
        elif Type == "Solar" : Player.Solar = True
        elif Type == "Celestial" : Player.Celestial = True

                                            # Only doing this if SavePreBakedAction is true
        #if Player.CurrentFight.SavePreBakedAction:
        #    fight = Player.CurrentFight
        #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15 , Target.ArcanumBuff.MultDPS)
        #    Target.PercentBuffHistory.append(history)

    ArcanumSpell = AstrologianSpell(0, False, Lock, 0, 0, 0, ApplyArcanum, [ArcanumRequirement])
    ArcanumSpell.TargetID = Target.playerID 
    ArcanumSpell.TargetPlayerObject = Target
    return ArcanumSpell

def Balance(target):
    returnSpell = Arcanum(target,"Solar", True)
    returnSpell.id = 4401
    return returnSpell

def Arrow(target):
    returnSpell = Arcanum(target,"Lunar", True)
    returnSpell.id = 4402
    return returnSpell

def Spear(target):
    returnSpell = Arcanum(target,"Celestial", True)
    returnSpell.id = 4403
    return returnSpell

def Bole(target):
    returnSpell = Arcanum(target,"Solar", False)
    returnSpell.id = 4404
    return returnSpell

def Ewer(target):
    returnSpell = Arcanum(target,"Lunar", False)
    returnSpell.id = 4405
    return returnSpell

def Spire(target):
    returnSpell = Arcanum(target,"Celestial", False)
    returnSpell.id = 4406
    return returnSpell

AstrologianAbility = {
3606 : Lightspeed,
25871 : Malefic,
16554 : Combust,
25872 : Gravity,
3590 : Draw,
3593 : Redraw, 
7443 : MinorArcana,
4401 : Balance,
4404 : Bole,
4402 : Arrow,
4405 : Ewer,
4403 : Spear,
4406 : Spire,
7444 : LordOfCrown,
7445 : LadyOfCrown,
25870 : Astrodyne,
16552 : Divination,
16559 : NeutralSect,
7439 : EarthlyStar,
3594 : Benefic,
3610 : Benefic2,
3595 : AspectedBenefic,
3600 : Helios,
3601 : AspectedHelios,
3612 : Synastry,
3613 : CollectiveUnconscious,
16553 : CelestialOpposition,
16556 : CelestialIntersection,
16558 : Horoscope,
16557 : Horoscope,
25873 : Exaltation,
25874 : Macrocosmos,
17 : Microcosmos,
3614 : EssentialDignity,
3603 : Ascend,
8324 : StellarDetonation
}


from ffxivcalc.Jobs.Base_Spell import buff, empty, buffHistory, buffPercentHistory
from ffxivcalc.Jobs.Ranged.Ranged_Spell import DancerSpell
import copy
#Requirement

def SambaRequirement(Player, Spell):
    return Player.SambaCD <= 0, Player.SambaCD

def CuringWaltzRequirement(Player, Spell):
    return Player.CuringWaltzCD <= 0, Player.CuringWaltzCD

def StandardStepRequirement(Player, Spell):
    return Player.StandardStepCD <= 0, Player.StandardStepCD

def DanceRequirement(Player, Spell):
    return Player.Dancing, -1

def StandardFinishRequirement(Player, Spell):
    #Will check how many step we have
    #Not done in apply since we want to acces the spell easily
    
                             # If no buff create one.
    if Player.StandardFinishTimer <= 0 :
        Player.StandardFinishBuff = copy.deepcopy(StandardFinishBuff)

    step = 0
    if Player.Emboite : step +=1
    if Player.Entrechat : step +=1
    if Player.Jete : step +=1
    if Player.Pirouette : step +=1

    if step == 1:
        Spell.Potency += 180
        Player.StandardFinishBuff.MultDPS = 1.02
    elif step > 1:
        Spell.Potency += 360
        Player.StandardFinishBuff.MultDPS = 1.05

    return Player.StandardFinish, -1
        
def TechnicalFinishRequirement(Player, Spell):
    #Will check how many step we have
    #Not done in apply since we want to acces the spell easily


    # Check if technical finish is already applied. If so we remove it
    # We check that by looking if check is in player's list
    if TechnicalFinishCheck in Player.EffectCDList and Player.TechnicalFinishBuff != None:
        # Is already applied. Remove buff
        Player.CurrentFight.Enemy.buffList.remove(Player.TechnicalFinishBuff)

    Player.TechnicalFinishBuff = copy.deepcopy(TechnicalFinishBuff)
    step = 0
    if Player.Emboite : step +=1
    if Player.Entrechat : step +=1
    if Player.Jete : step +=1
    if Player.Pirouette : step +=1

    if step == 1 :
        Spell.Potency += 190
        Player.TechnicalFinishBuff.MultDPS = 1.01
    elif step == 2 :
        Spell.Potency += 370
        Player.TechnicalFinishBuff.MultDPS = 1.02
    elif step == 3 :
        Spell.Potency += 550
        Player.TechnicalFinishBuff.MultDPS = 1.03
    elif step == 4 :
        Spell.Potency += 850
        Player.TechnicalFinishBuff.MultDPS = 1.05

    return Player.TechnicalFinish, -1

def TechnicalStepRequirement(Player, Spell):
    return Player.TechnicalStepCD <= 0, Player.TechnicalStepCD

def DevilmentRequirement(Player, Spell):
    return Player.DevilmentCD <= 0, Player.DevilmentCD

def TillanaRequirement(Player, Spell):
    return Player.FlourishingFinish, -1

def StarfallDanceRequirement(Player, Spell):
    return Player.FlourishingStarfall, -1

def FlourishRequirement(Player, Spell):
    return Player.FlourishCD <= 0, Player.FlourishCD

def FanDance4Requirement(Player, Spell):
    return Player.FourfoldFan, -1

def FanDance3Requirement(Player, Spell):
    return Player.ThreefoldFan, -1

def FanDance1Requirement(Player, Spell):
    return Player.MaxFourfoldFeather > 0, -1

def FountainFallRequirement(Player, Spell):
    return Player.SilkenFlow or Player.FlourishingFlow, -1

def ReverseCascadeRequirement(Player, Spell):
    return Player.SilkenSymettry or Player.FlourishingSymettry, -1

def SaberDanceRequirement(Player, Spell):
    return Player.MaxEspritGauge >= 50, -1

def ClosedPositionRequirement(Player, Spell):
    return Player.ClosedPositionCD <= 0, Player.ClosedPositionCD

def ImprovisedFinishRequirement(Player, Spell):
    return Player.Improvising, -1

#Apply

def ApplySamba(Player, Enemy):
    Player.SambaCD = 90

def ApplyImprovisedFinish(Player, Enemy):
    Player.Improvising = False

def ApplyCuringWaltz(Player, Enemy):
    Player.CuringWaltzCD = 60

def ApplyWindmill(Player, Enemy):
    if not (WindmillCombo in Player.EffectList) : Player.EffectList.append(WindmillCombo)

def ApplyCascade(Player, Enemy):
    #Cascade has 50% chance to grant SikenSymettry, so we will keep track of the expected number we should have
    Player.ExpectedSilkenSymettry += 0.5 #adding to expected
    Player.SilkenSymettry = True    #Assuming it is true

    #Adding Combo Action
    if not (CascadeComboEffect in Player.EffectList) : Player.EffectList.append(CascadeComboEffect)


def ApplyStandardStep(Player, Enemy):
    Player.Dancing = True
    Player.StandardStepCD = 30
    Player.StandardFinish = True #Ready StandardFinish

def ApplyStandardFinish(Player, Enemy):
    #The MultBonus has already been computed from Requirement, so we just apply it to the dancer
    #and the dance partner

    #The check will be done on the dancer only
    if Player.StandardFinishTimer <= 0: 
        Player.EffectCDList.append(StandardFinishCheck)
        Player.buffList.append(Player.StandardFinishBuff) #Bonus DPS
        if Player.DancePartner != None : Player.DancePartner.buffList.append(Player.StandardFinishBuff) #Dance Partner Bonus
    
    Player.StandardFinishTimer = 60

    Player.Emboite = False
    Player.Entrechat = False
    Player.Jete = False
    Player.Pirouette = False
    #Reseting Dance move

def ApplyTechnicalStep(Player, Enemy):
    Player.Dancing = True
    Player.TechnicalStepCD = 120
    Player.TechnicalFinish = True #Ready TechnicalFinish

def ApplyTechnicalFinish(Player, Enemy):
    #The MultBonus has already been computed from Requirement, so we just apply it to the dancer
    #and the dance partner

    # Add only if not present.
    if not (TechnicalFinishCheck in Player.EffectCDList):
        Player.EffectCDList.append(TechnicalFinishCheck)
        
    Enemy.buffList.append(Player.TechnicalFinishBuff) #Since party wide, applies on enemy


    Player.TechnicalFinishTimer = 20
    Player.Emboite = False
    Player.Entrechat = False
    Player.Jete = False
    Player.Pirouette = False
    #Reseting Dance move

    Player.FlourishingFinish = True #Enables Tillana

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , Player.TechnicalFinishBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyEmboite(Player, Enemy):
    Player.Emboite = True

def ApplyEntrechat(Player, Enemy):
    Player.Entrechat = True

def ApplyJete(Player, Enemy):
    Player.Jete = True

def ApplyPirouette(Player, Enemy):
    Player.Pirouette = True

def ApplyDevilment(Player, Enemy):
    Player.FlourishingStarfall = True #Enables StarfallDance

    Player.DevilmentTimer = 20
    Player.DevilmentCD = 120
    #Give buff

    Player.CritRateBonus += 0.2
    Player.DHRateBonus += 0.2

    if Player.DancePartner != None : 
        Player.DancePartner.CritRateBonus += 0.2
        Player.DancePartner.DHRateBonus += 0.2

    Player.EffectCDList.append(DevilmentCheck)

def ApplyTillana(Player, Enemy):
    Player.FlourishingFinish = False 

    if Player.StandardFinishTimer <= 0:
        Player.StandardFinishBuff = copy.deepcopy(StandardFinishBuff)

    #We have to apply or reapply StandardFinish with a bonus of 5%, so reset, set bonus to 5% and apply
    #if Player.StandardFinishBuff != None : Player.buffList.remove(Player.StandardFinishBuff) #Reset DPSBonus
    #if Player.DancePartner != None and Player.StandardFinishBuff in Player.DancePartner.buffList : Player.DancePartner.buffList.remove(Player.StandardFinishBuff) #Reset DPSBonus
    
    Player.StandardFinishBuff.MultDPS = 1.05

    ApplyStandardFinish(Player, Enemy) #Will give StandardFinish with a bonus of 5%

def ApplyStarfallDance(Player, Enemy):
    Player.NextDirectCrit = True
    Player.FlourishingStarfall = False

def ApplyFlourish(Player, Enemy):
    Player.FlourishCD = 60
    Player.FlourishingSymettry = True
    Player.FlourishingFlow = True
    Player.ThreefoldFan = True
    Player.FourfoldFan = True
    Player.ExpectedThreefoldFan +=1

def ApplyFanDance4(Player, Enemy):
    Player.FourfoldFan = False

def ApplyFanDance3(Player, Enemy):
    Player.ThreefoldFan = False
    Player.UsedThreefoldFan += 1

def ApplyFanDance1(Player, Enemy):
    Player.ThreefoldFan = True #Assume it happend
    Player.ExpectedThreefoldFan += 0.5 #add to expected

    Player.MaxFourfoldFeather -= 1
    Player.UsedFourfoldFeather += 1

def ApplyFountainFall(Player, Enemy):
    Player.MaxFourfoldFeather = min(4, Player.MaxFourfoldFeather + 1)
    Player.ExpectedFourfoldFeather += 0.5
    #Stats
    if not Player.FlourishingFlow: #If not from flourish
        Player.SilkenFlow = False
        Player.UsedSilkenFlow += 1
    else: #From flourish
        Player.FlourishingFlow = False #Not RNG

def ApplyReverseCascade(Player, Enemy):

    Player.MaxFourfoldFeather = min(4, Player.MaxFourfoldFeather + 1) 
    Player.ExpectedFourfoldFeather += 0.5 #Create
    if not Player.FlourishingSymettry: #Then its by SilkenSymettry
        Player.UsedSilkenSymettry += 1 #Need
        Player.SilkenSymettry = False
    else:
        Player.FlourishingSymettry = False
        #Since Flourish isn't RNG, this one does not keep track of anything

def ApplySaberDance(Player, Enemy):
    Player.MaxEspritGauge -= 50

def ApplyEnding(Player, Enemy):
    if Player.StandardFinishTimer > 0: #Remove StandardFinish
        Player.DancePartner.buffList.remove(Player.StandardFinishBuff)
    if Player.DevilmentTimer > 0: #Remove Devilment
        Player.DancePartner.CritRateBonus -= 0.2
        Player.DancePartner.DHRateBonus -= 0.2

    Player.DancePartner = None #Removing DancePartner

#Effect

def WindmillCombo(Player, Spell):
    if Spell.id == Bladeshower.id:
        Spell.Potency += 40
        Spell.Effect = Fountain.Effect #Giving Effect since we have made the combo

def EspritEffect(Player, Spell):
    if Spell.id == Fountain.id or Spell.id == FountainFall.id or Spell.id == Cascade.id or Spell.id == ReverseCascade.id:
        Player.MaxEspritGauge = min(100, Player.MaxEspritGauge + 5)

def CascadeComboEffect(Player, Spell):
    if Spell.id == Fountain.id:
        Spell.Potency += 180
        
        Player.ExpectedSilkenFlow += 0.5
        Player.SilkenFlow = True
        Player.EffectToRemove.append(CascadeComboEffect)

#Check

def DevilmentCheck(Player, Enemy):
    if Player.DevilmentTimer <= 0:
        Player.CritRateBonus -= 0.2
        Player.DHRateBonus -= 0.2
        if Player.DancePartner != None : 
            Player.DancePartner.CritRateBonus -= 0.2
            Player.DancePartner.DHRateBonus -= 0.2
        Player.EffectToRemove.append(DevilmentCheck)

def StandardFinishCheck(Player, Enemy):
    if Player.StandardFinishTimer <= 0:
        Player.buffList.remove(Player.StandardFinishBuff)
        if Player.DancePartner != None : Player.DancePartner.buffList.remove(Player.StandardFinishBuff)
        Player.EffectToRemove.append(StandardFinishCheck)
        Player.StandardFinishBuff = None

def TechnicalFinishCheck(Player, Enemy):
    if Player.TechnicalFinishTimer <= 0:
        Enemy.buffList.remove(Player.TechnicalFinishBuff)
        Player.EffectToRemove.append(TechnicalFinishCheck)
        Player.TechnicalFinishBuff = None


#GCD
StarfallDance = DancerSpell(25792, True, 2.5, 600, ApplyStarfallDance, [StarfallDanceRequirement], True, type = 2)
#ComboAction
Cascade = DancerSpell(15989, True, 2.5, 220,ApplyCascade, [], True, type = 2)
Fountain = DancerSpell(15990, True, 2.5, 100, empty, [], True, type = 2)
FountainFall = DancerSpell(15992, True, 2.5, 340, ApplyFountainFall, [FountainFallRequirement], True, type = 2)
ReverseCascade = DancerSpell(15991, True, 2.5, 280, ApplyReverseCascade, [ReverseCascadeRequirement], True, type = 2)
SaberDance = DancerSpell(16005, True, 2.5, 480, ApplySaberDance, [SaberDanceRequirement], True, type = 2)
#AOE Combo action
Windmill = DancerSpell(15993, True, 2.5, 100, ApplyWindmill, [], True , type = 2) #AOE Version of Cascade
RisingWindmill = DancerSpell(15995, True, 2.5, 140, ApplyReverseCascade, [ReverseCascadeRequirement], True, type = 2) #AOE version of Reverse Cascade
Bladeshower = DancerSpell(15994, True, 2.5, 100, empty, [], True, type = 2)
Bloodshower = DancerSpell(15996, True, 2.5, 180, ApplyFountainFall, [FountainFallRequirement], True, type = 2)
#Dance
StandardStep = DancerSpell(15997, True, 1.5, 0, ApplyStandardStep, [StandardStepRequirement], True, type = 2)
StandardFinish = DancerSpell(16192, True, 1.5, 360, ApplyStandardFinish, [StandardFinishRequirement], True, type = 2)
TechnicalStep = DancerSpell(15998, True, 1.5, 0, ApplyTechnicalStep, [TechnicalStepRequirement], True, type = 2)
TechnicalFinish = DancerSpell(16004, True, 1.5, 350, ApplyTechnicalFinish, [TechnicalFinishRequirement], True, type = 2)

def Improvisation(time): #Function, since we need to know for how long we are doing it

    def ImprovisationRequirement(Player, Spell):
        return Player.ImprovisationCD <= 0, Player.ImprovisationCD

    def ApplyImprovisation(Player, Enemy):
        Player.ImprovisationCD = 0
        Player.Improvising = True
    
    return DancerSpell(16014, True, time, time, ApplyImprovisation, [ImprovisationRequirement], False)

ImprovisedFinish = DancerSpell(25789, True, 1.5, 0, ApplyImprovisedFinish, [ImprovisedFinishRequirement], False)
Tillana = DancerSpell(25790, True, 1.5, 360, ApplyTillana, [TillanaRequirement], True, type = 2)
#Dance Move
Emboite = DancerSpell(15999, True, 1, 0, ApplyEmboite, [DanceRequirement], True)
Entrechat = DancerSpell(16000, True, 1, 0, ApplyEntrechat, [DanceRequirement], True)
Jete = DancerSpell(16001, True, 1, 0, ApplyJete, [DanceRequirement], True)
Pirouette = DancerSpell(16002, True, 1, 0, ApplyPirouette, [DanceRequirement], True)


#oGCD
Devilment = DancerSpell(16011, False, 0, 0, ApplyDevilment, [DevilmentRequirement], False)
Flourish = DancerSpell(16013, False, 0, 0, ApplyFlourish, [FlourishRequirement], False)
FanDance4 = DancerSpell(25791, False, 0, 300, ApplyFanDance4, [FanDance4Requirement], False)
FanDance3 = DancerSpell(16009, False, 0, 200, ApplyFanDance3, [FanDance3Requirement], False)
FanDance2 = DancerSpell(16008, False, 0, 100, ApplyFanDance1, [FanDance1Requirement], False) #AOE Version of FanDance1
FanDance1 = DancerSpell(16007, False, 0, 150, ApplyFanDance1, [FanDance1Requirement], False)
EnAvant = DancerSpell(16010, False, 0, 0, empty, [], False) #No requirement, spam that shit
CuringWaltz = DancerSpell(16015, False, 0, 0, ApplyCuringWaltz, [CuringWaltzRequirement], False)
Samba = DancerSpell(16012, False, 0, 0, ApplySamba, [SambaRequirement], False)
#buff
TechnicalFinishBuff = buff(1,name="TF") #We will adapt the buff depending on number of steps at casting
StandardFinishBuff = buff(1,name="SF") #We will adapt the buff depending on number of steps at casting

#Dance Partner
Ending = DancerSpell(18073, False, 0, 0, ApplyEnding, [], False)

def ClosedPosition(Partner, InFight="True"):
    def ApplyClosedPosition(Player, Enemy):
                             # Updating target ID once the action has been done.
        ClosedPositionSpell.TargetID = Partner.playerID

        if InFight : Player.ClosedPositionCD = 30 #Only update CD if in Fight, since we can dance partner before begins

        Player.DancePartner = Partner #New Partner
        if Player.StandardFinishTimer > 0: #Add StandardFinish
            Player.DancePartner.buffList.append(Player.StandardFinishBuff)
        if Player.DevilmentTimer > 0: #Add Devilment
            Player.DancePartner.CritRateBonus += 0.2
            Player.DancePartner.DHRateBonus += 0.2


        #Will have to switch buff
        # If change this ID make sure to change in deepCopyFight too.
    ClosedPositionSpell = DancerSpell(16006, False, 0, 0, ApplyClosedPosition, [ClosedPositionRequirement], False)
    ClosedPositionSpell.TargetID = Partner.playerID
    ClosedPositionSpell.TargetPlayerObject = Partner
    return ClosedPositionSpell

DancerAbility = {
25792 : StarfallDance, 
25791 : FanDance4, 
16015 : CuringWaltz, 
16014 : Improvisation(2), 
16013 : Flourish, 
16012 : Samba, 
16011 : Devilment, 
16010 : EnAvant, 
16009 : FanDance3, 
16008 : FanDance2, 
16007 : FanDance1, 
16006 : ClosedPosition, 
16005 : SaberDance, 
15997 : StandardStep, 
15998 : TechnicalStep,
15996 : Bloodshower, 
15995 : RisingWindmill, 
15994 : Bladeshower, 
15993 : Windmill, 
15992 : FountainFall, 
15991 : ReverseCascade, 
15990 : Fountain, 
15989 : Cascade, 
16004 : TechnicalFinish, 
16193 : TechnicalFinish, 
16194 : TechnicalFinish, 
16195 : TechnicalFinish,
16196 : TechnicalFinish,
16192 : StandardFinish, 
25789 : ImprovisedFinish , 
18073 : Ending, 
25790 : Tillana, 
15999 : Emboite, 
16000 : Entrechat, 
16001 : Jete, 
16002 : Pirouette 
}


"""
This file contains the class SimulationRecord. This class contains a list of events to be recapped. 
For now events are defined as an action that deals damage. It will be recorded with some information.
the SimulationRecord object will then be able to output an array that contains the information of the whole fight in a clear way.

A SimulationRecord contains a list of page which are defined as individual events. There are multiple subclasses of page each defining a specific
event. maybe not? have to think. Ima keep it simple for now

This is supposed to be an alternative to logs. Making something that has still a good amount of information but is easier to read.
"""

import matplotlib.pyplot as plt

class page:
    """
    This contains the information of one individual event.
    Potency (int) : Potency of the action
    Damage (int) : Total Damage of the action
    TimeStamp (float) : TimeStamp of the action (when it applies damage)
    PercentBuffList (list[buff]) : List of Percent Buff.
    DHBuffList (list) : List of DH buff
    CritBuffList (list) : List of Crit buff
    """

    def __init__(self):
        self.Name = "Unnamed"
        self.Potency = 0
        self.Damage = 0
        self.TimeStamp = 0.0
        self.autoCrit = False
        self.autoDH = False
        self.PercentBuffList = []
        self.DHBuffList = []
        self.CritBuffList = []
        self.playerID = 0
        self.hasPotion = False
        self.isAuto = False
        self.isDOT = False

    def setPlayerID(self,newID : int ):
        self.playerID = newID

    def setName(self, name : str):
        self.Name = name

    def setPotency(self, newPotency : int):
        self.Potency = newPotency

    def setDamage(self, newDamage : int):
        self.Damage = newDamage

    def setTimeStamp(self, newTimeStamp : float):
        self.TimeStamp = newTimeStamp

    def setAutoCrit(self, value : bool):
        self.autoCrit = value

    def setAutoDH(self, value : bool):
        self.autoDH = value

    def addPercentBuff(self, newEntry):
        self.PercentBuffList.append(newEntry)

    def addDHBuffList(self, newEntry):
        self.DHBuffList.append(newEntry)

    def addCritBuffList(self, newEntry):
        self.CritBuffList.append(newEntry)

    def setHasPotion(self):
        """This function sets the value of the field hasPotion to True
        """
        self.hasPotion = True
    
    def setIsDOT(self):
        """sets self.isDOT to true
        """
        self.isDOT = True

    def setIsAuto(self):
        """sets self.isAuto to true
        """
        self.isAuto = True

    def isValid(self,startTime : float, endTime : float, trackAutos : bool, trackDOTs : bool, idList) -> bool:
        """This function returns weither the page would be shown under these restrictions. 
        See SimulationRecord.getRecordLength for a description of the arguments.
        """
        return self.TimeStamp >= startTime and self.TimeStamp <= endTime and (self.isAuto and trackAutos or (not self.isAuto)) and (self.isDOT and trackDOTs  or (not self.isDOT)) and self.playerID in idList

    def __str__(self) -> str:

        percentBuffStr = ""
        dhBuffStr = ""
        critBuffStr = ""
        
        for buffs in self.PercentBuffList:
            percentBuffStr += str(buffs)
        for buffs in self.DHBuffList:
            dhBuffStr += buffs[0] + "(" + str(int(buffs[1]*100)) + "%)"
        for buffs in self.CritBuffList:
            critBuffStr += buffs[0] + "(" + str(int(buffs[1]*100)) + "%)"

        crititalDH = ""

        if self.autoCrit : crititalDH += "!"
        if self.autoDH : crititalDH += "!"

        if self.hasPotion : crititalDH += " Pot"

        return self.Name + " ; " + str(self.Potency) + " ; " + str(self.Damage) + " " + crititalDH + " ; " + " ; " + str(self.TimeStamp) + " ; " + percentBuffStr + " ; " + dhBuffStr + " ; " + critBuffStr

class SimulationRecord:

    def __init__(self):
        self.pageList = []
        self.idList = []

    def addPage(self, newPage : page):
        self.pageList.append(newPage)

    def __str__(self):
        rString = ""
        specId = len(self.idList) == 0

        for pages in self.pageList:
            if specId or (pages.playerID in self.idList) : rString += str(pages) + "\n"

        return rString

    def getRecordLength(self, startTime : float, endTime : float, trackAutos : bool, trackDOTs : bool, idList) -> int:
        """This function returns the length of the record under the given restrictions

        Args:
            startTime (float): Record only shows pages with time after this.
            endTime (float): Record only shows pages with time before this.
            trackAutos (bool): If true the record will show autos events.
            trackDOTs (bool): If true the record will show DOTs events.
            idList (_type_): Player's id that the record will show.
        """
        nRows = 0

        for page in self.pageList:
            if page.isValid(startTime, endTime, trackAutos , trackDOTs, idList): nRows+=1

        return nRows

    def saveRecordText(self, idList : list = [], path : str = 'SimulationRecord.txt', startTime : float = 0, endTime : float = 99999, trackAutos : bool = True,
                       trackDOTs : bool = True, customizeRecord : bool = False):
        """
        This function saves the record as a text file. Read saveRecord for info on arguments.
        """

        self.idList = idList

        rString = ""
        specId = len(self.idList) == 0

        for pages in self.pageList:
            if not customizeRecord or pages.isValid(startTime, endTime, trackAutos, trackDOTs,idList): rString += str(pages) + '\n'

        f = open(path, "w")
        f.write(rString)
        f.close()

    def saveRecord(self,customizeRecord : bool = False, startTime : float = 0, endTime : float = 99999, trackAutos : bool = True, 
                   trackDOTs : bool = True, idList : list = [],saveAsPDF : bool =True):
        """
        This function saves the record as a plot (pdf). See getRecordLength for a description of the arguments.

        idList : list[int] -> If non empty will limit the record's output to player that match the id inside the list
        saveAsPDF : bool -> If true saves as a PDF
        customizeRecord : bool -> If true will limit pages to the given restrictions.
        """

        colName = ["Name", "Potency", "Damage", "", "Time","Buff", "DH Buff", "Crit Buff"]
        haList = ["left","right","right","center","right","left","center","center"]

        posOffSet = [1.3,0.8,0.1,0.5,0.2,4.8,2]
                             # Generating pos list
        curPos = 0
        pos = [0]
        for offset in posOffSet:
            curPos += offset
            pos.append(curPos)


                             # Will go through pageList and only append the pages that we want
        newPageList = [] 

        if not customizeRecord:
            newPageList = self.pageList
        else:
            for page in self.pageList:
                if page.isValid(startTime, endTime, trackAutos, trackDOTs, idList):
                    newPageList.append(page)


        nrows = len(newPageList)
        ncols = len(colName)
                             # This works well. Found by just using a configuration that looked good
                             # and using the ratio of nrows to height.
        height = ((160/538) * nrows)

        fig = plt.figure(figsize=(14,height), dpi=700)
        print("nrows : " + str(nrows))
        #if nrows > 500 : return fig
        ax = plt.subplot(111)
        ax.set_xlim(0, int(1.5*ncols))
        ax.set_ylim(0, 5*(nrows+1))
        ax.set_axis_off()
        offset = 0
        size = 6

        for y in range(nrows-1,-1,-1):
            curPercentCount = 0
            percentBuffStr = ""
            dhBuffStr = ""
            critBuffStr = ""

            for buffs in newPageList[nrows-y-1].PercentBuffList:
                curPercentCount += 1
                percentBuffStr += str(buffs) + ("\n" if curPercentCount%4 == 0 else " ")

            if newPageList[nrows-y-1].hasPotion : percentBuffStr += "Potion "

            for buffs in newPageList[nrows-y-1].DHBuffList:
                dhBuffStr += buffs[0] + "(" + str(int(buffs[1]*100)) + "%) "
            for buffs in newPageList[nrows-y-1].CritBuffList:
                critBuffStr += buffs[0] + "(" + str(int(buffs[1]*100)) + "%) "

                             # Removing last character which is a \n
            percentBuffStr = percentBuffStr[:-1]
            dhBuffStr = dhBuffStr[:-1]
            critBuffStr = critBuffStr[:-1]
            yPos = 5*(y+1)-2
            
            ax.annotate(
                xy=(pos[0],yPos),
                text=newPageList[nrows-y-1].Name,
                ha=haList[0],
                fontsize=size+1
            )
            ax.annotate(
                xy=(pos[1],yPos),
                text=newPageList[nrows-y-1].Potency,
                ha=haList[1],
                fontsize=size+1
            )
            ax.annotate(
                xy=(pos[2],yPos),
                text=newPageList[nrows-y-1].Damage,
                ha=haList[2],
                fontsize=size+1
            )
            crititalDH = ""
            if newPageList[nrows-y-1].autoCrit : crititalDH += "!"
            if newPageList[nrows-y-1].autoDH : crititalDH += "!"
            ax.annotate(
                xy=(pos[3],yPos),
                text=crititalDH,
                ha=haList[3],
                fontsize=8,
                weight="bold"
            )
            ax.annotate(
                xy=(pos[4],yPos),
                text=newPageList[nrows-y-1].TimeStamp,
                ha=haList[4],
                fontsize=size
            )
            ax.annotate(
                xy=(pos[5],yPos),
                text=percentBuffStr,
                ha=haList[5],
                fontsize=size if curPercentCount <= 4 else 5
            )
            ax.annotate(
                xy=(pos[6],yPos),
                text=dhBuffStr,
                ha=haList[6],
                fontsize=6
            )
            ax.annotate(
                xy=(pos[7],yPos),
                text=critBuffStr,
                ha=haList[7],
                fontsize=6
            )

        for j, name in enumerate(colName):
            ax.annotate(
                xy=(pos[j], 5*(nrows+1)),
                text=name,
                weight='bold',
                ha=haList[j]
            )

        ax.plot([ax.get_xlim()[0], ax.get_xlim()[1]], [5*(nrows+1)-1, 5*(nrows+1)-1], lw=1.5, color='black', marker='', zorder=4)
        ax.plot([ax.get_xlim()[0], ax.get_xlim()[1]], [0, 0], lw=1.5, color='black', marker='', zorder=4)
        for x in range(1, 5*(nrows),5):
            ax.plot([ax.get_xlim()[0], ax.get_xlim()[1]], [x+1, x+1], lw=1.15, color='gray', ls=':', zorder=3 , marker='')
            
        if saveAsPDF:
            try:
                plt.savefig(
                'SimulationRecord.pdf',
                dpi=700,
                bbox_inches='tight'
                )
            except:
                print("An error happened while trying to export the Simulation Record. This could be due to a too large size.")

            

        return fig





__version__ = '0.8.940'

"""
Update history :

0.9.940:
    - Oups, forgot to ever implement Army Ethos/Army Muse effect on Bard. This is now fixed.
    - Other bug fixes.

0.8.930:
    - Added Player.getPlayerPrePullTime() which returns the length of the prepull (time before first damage action).
    - Added Fight.getPlayerPrepullLength() which returns the prepull length of all players
    - added Fight.syncPlayerPrePull() which syncs all player's prepull so they all do damage at the same time.
    - Added test suite 'Prepull length test suite' that tests the Player.GetPlayerPrePullTime() function.
    - Removed access to ffxvivcalc.API, ffxivcalc.codeParser and ffxivcalc.Request.FFLogsAPIRequest modules (outdated modules).
    - Other minor bug fix.

0.8.925:
    - getSingleFightData() can take 'max_time' input.
    - _aura_prepull_logic() is called in getSingleFightData() instead of RestoreFightObject. So prepull are now put directly into JSON file exported.
    - Fixed some other issues

0.8.920:
    - Added code to import fights from FFlogs (big thanks to Apollo for the base code).
    - Added FFLogs test suite.
    - Discontinued FFLogsAPIRequest
    - Some bug fixes and other smaller improvements.

0.8.911:
    - Fixed Bard's Mine issues
    - Added functionality to filter the record under some restrictions.
    - Fixed technical finish issue where a double cast would make the simulation crash
    - Fixed GivenHealBuff issue
    - Graph from the simulation now save as png instead of opening up.

0.8.910:
    - Fixed some 'computeActionTimer()' issues
    - Added Limit break actions (they do no damage but should be accurate on cast time/animation lock)
    - Fixed some wrong Ids issue
    - Fixed Huraijin ninja issue
    - Added Limit break cast time tests for SpS/SkS interactions.

0.8.901 :
    - Fixed some 'computeActionTimer()' issues on Samurai with Meikyo/Gekko interaction, an issue where only having 1 GCD would return wrong estimated time stamp.

0.8.900 :
    - Added 'computeActionTimer()' which returns an estimate of a player's timeStamp/DOT timer/buf timer.
      This function can be called on a player object and will return a dictionnary containing the estimates.
    - Added tests to test 'computeActionTimer' functionality.
    - Fixed mitigation amount issue (higher than 1 values) (PR by Apollo)
    - Fixed Samurai Higanbana DOT potency (was 60, fixed to 45) (PR by Apollo)
    - Moved some gcdTiming computation to the Player object (see Fight.py)
    - Removed lock in various places

0.8.800 :
    - Made changes to work with web app
    - Added ways to remove pB or redirect the output
    - Fixed etro import issues
    - BiSSolver now makes a deepcopy of the gear space so it can be reused
    - Fixed pb issues, fixed fight import issues.
    - Added 'ShowProgress' option to 'SimulateFight()' that lets enables the print of the loading bar
    - Various other bug fixes

0.8.700 :
    - Made PyPi package.

0.8.60 :
    - Added more accurate AA, weapon delay, haste affecting delay.
    - Fixed DOT/AA issues
    - Added AA delay, GCD timer and more DOT tests.
    - Added cli "tests" command to run test suites.
    

"""



# This will contain all action with their IDs in the game. It will map from ids -> name or name -> ids
from enum import IntEnum
import logging
main_logging = logging.getLogger("ffxivcalc")
action_logging = main_logging.getChild("ActionEnum")

from ffxivcalc.Jobs.PlayerEnum import JobEnum, RoleEnum

class ActionEnum(IntEnum):
    # Parent enum class for all other enums. Will have
    # the two functions.

    @classmethod
    def name_for_id(cls, id : int) -> str:
        # maps from id -> name
        if id in cls.__members__.values():
            return cls(id).name
        return 'Unknown'

    @classmethod
    def id_for_name(cls, name : str) -> int:
        # maps from name -> id
        if name in cls.__members__.keys():
            return cls[name].value
        return -1 # Evaluated as Unknown

# Caster

class CasterActions(ActionEnum):
    Swiftcast = 7561
    LucidDreaming = 7562
    Surecast = 7559
    Addle = 7560
    Sleep = 25880
    Potion = -2
    WaitAbility = 212
    
    # Limit Break 
    CasterLB1 = 1111
    CasterLB2 = 1112
    CasterLB3 = 1113


# BlackMage

class BlackMageActions(ActionEnum):
    Paradox = 25797
    FireI = 141
    FireII = 147
    FireIII = 152
    FireIV = 3577
    Despair = 16505
    Flare = 162
    HighFireII = 25794
    BlizzardI = 142
    BlizzardII = 25793
    BlizzardIII = 154
    BlizzardIV = 3576
    Freeze = 159
    HighBlizzardII = 25795
    UmbralSoul = 16506
    ThunderIII = 153
    ThunderIV = 7420
    Transpose = 149
    AetherialManipulation = 155
    Scathe = 156
    Manaward = 157
    Manafont = 158
    LeyLines = 3573
    Sharpcast = 3574
    BetweenTheLines = 7419
    Triplecast = 7421
    Foul = 7422
    Xenoglossy = 16507
    Amplifier = 25796

    # DOT
    Thunder4Dot = -40
    Thunder3DOT = -21


# RedMage

class RedMageActions(ActionEnum):
    Riposte = 7504
    Corps = 7506
    Verfire = 7510
    Verstone = 7511
    Zwerchhau = 7512
    Moulinet = 7513
    Vercure = 7514
    Redoublement = 7516
    Fleche = 7517
    Acceleration = 7518
    Contre = 7519
    Embolden = 7520
    Manafication = 7521
    Verraise = 7523
    Jolt = 7524
    Verthunder = 25855
    Verareo = 25856
    Impact = 16526
    Engagement = 16527
    Reprise = 16529
    MagickBarrier = 25857
    EnchantedRiposte = 7527
    EnchantedZwerchhau = 7528
    EnchantedRedoublement = 7529
    EnchantedMoulinet = 7530
    EnchantedReprise = 16528
    Resolution = 25858
    Verholy = 7526
    Verflare = 7525
    Scorch = 16530


# Summoner

class SummonerActions(ActionEnum):
    RuinIII = 3579
    RuinIV = 7426
    TryDisaster = 25826
    SearingLight = 25801
    SummonBahamut = 7427
    AstralImpulse = 25820
    AstralFlare = 25821
    Deathflare = 3582
    EnkindleBahamut = 7429
    SummonPhoenix = 25831
    FountainOfFire = 16514
    BrandOfPurgatory = 16515
    EnkindlePhoenix = 16516
    Rekindle = 25830
    Ifrit = 25838
    Ruby = 25823
    RubyCatastrophe = 25832
    Cyclone = 25835
    Strike = 25885
    Titan = 25839
    Topaz = 25824
    TopazCatastrophe = 25833
    Mountain = 25836
    Garuda = 25840
    Emerald = 25825
    EmeraldCatastrophe = 25834
    Slipstream = 25837
    EnergyDrain = 16508
    EnergySyphon = 16510
    Fester = 181
    Painflare = 3578
    Physick = 16230
    Resurrection = 173
    Summon = 7427
    RadiantAegis = 25799

    # Pet action
    SummonEnkindle = -17
    ApplySummonAuto = -115
    BahamutAuto = -14
    PhoinixAuto = -15
    SlipstreamDOT = 100


# Healer

class HealerActions(ActionEnum):
    Repose = 16560
    Esuna = 7568
    Rescue = 7571
    Swiftcast = 7561
    LucidDreaming = 7562
    Surecast = 7559
    Potion = -2
    WaitAbility = 212

    # Limit Break
    HealerLB1 = 1111
    HealerLB2 = 1112
    HealerLB3 = 1113


# Astrologian

class AstrologianActions(ActionEnum):
    EssentialDignity = 3614
    StellarDetonation = 8324
    Ascend = 3603
    Lightspeed = 3606
    Malefic = 25871
    Combust = 16554
    Gravity = 25872
    Draw = 3590
    Redraw = 3593
    MinorArcana = 7443
    Balance = 4401
    Bole = 4404
    Arrow = 4402
    Ewer = 4405
    Spear = 4403
    Spire = 4406
    CelestialArcanum = 4403
    LordOfCrown = 7444
    LadyOfCrown = 7445
    Astrodyne = 25870
    Divination = 16552
    NeutralSect = 16559
    EarthlyStar = 7439
    Benefic = 3594
    BeneficII = 3610
    AspectedBenefic = 3595
    Helios = 3600
    AspectedHelios = 3601
    Synastry = 3612
    CollectiveUnconscious = 3613
    CelestialOpposition = 16553
    CelestialIntersection = 16556
    Horoscope = 16558
    Exaltation = 25873
    Macrocosmos = 25874
    Microcosmos = 17

    #DOT
    CombustDOT = -12


# Sage

class SageActions(ActionEnum):
    Egeiro = 24287
    Dosis = 24312
    EukrasianDosis = 24314
    Toxikon = 24316
    Dyskrasia = 24315
    Phlegma = 24313
    Kardia = 24285
    Eukrasia = 24290
    Diagnosis = 24284
    EukrasianDiagnosis = 24291
    Prognosis = 24286
    EukrasianPrognosis = 24292
    Physis = 24302
    Soteria = 24294
    Icarus = 24295
    Druochole = 24296
    Kerachole = 24298
    Ixochole = 24299
    Zoe = 24300
    Pepsis = 24301
    Taurochole = 24303
    Haima = 24305
    Rhizomata = 24309
    Holos = 24310
    Panhaima = 24311
    Krasis = 24317
    Pneuma = 24318

    # DOT
    EnkrasianDOT = -12

# Scholar

class ScholarActions(ActionEnum):
    Resurrection = 173
    Aetherflow = 166
    EnergyDrain = 167
    Adloquium = 185
    Succor = 186
    SacredSoil = 188
    Lustrate = 189
    Physick = 190
    Indomitability = 3583
    DeploymentTactic = 3585
    EmergencyTactic = 3586
    Dissipation = 3587
    Excogitation = 7434
    ChainStratagem = 7436
    Aetherpact = 7437
    WhisperingDawn = 16537
    FeyIllumination = 16538
    Recitation = 16542
    FeyBlessing = 16543
    SummonSeraph = 16545
    SummonEos = 16545
    Expedient = 25868
    Protraction = 25867
    ArtOfWar = 25866
    BroilIV = 25865
    RuinII = 17870
    Biolysis = 16540
    Consolation = 16546
    Raise = 173
    DissolveUnion = 7869

    # DOT
    BiolysisDOT = -4



# Whitemage

class WhiteMageActions(ActionEnum):
    Glare = 25859
    Dia = 16532
    Holy = 25860
    AfflatusMisery = 16535
    Assize = 3571
    Cure = 120
    CureII = 135
    CureIII = 131
    Medica = 124
    MedicaII = 133
    Regen = 137
    Tetragrammaton = 3570
    Asylum = 3569
    DivineBenison = 7432
    Benediction = 140
    AfflatusSolace = 16531
    AfflatusRapture = 16534
    PresenceOfMind = 136
    ThinAir = 7430
    Temperance = 16536
    Aquaveil = 25861
    Bell = 25862
    PlenaryIndulgence = 7433
    Raise = 125

    # DOT
    DiaDOT = -5

# Tank

class TankActions(ActionEnum):
    Rampart = 7531
    Reprisal = 7535
    ArmLength = 7548
    LowBlow = 7540
    Provoke = 7533
    Shirk = 7537
    RoyalGuard = 16142
    ReleaseRoyalGuard = 16142
    Grit = 3629
    ReleaseGrit = 3629
    IronWill = 28
    ReleaseIronWill = 28
    Defiance = 48
    ReleaseDefiance = 48
    Potion = -2
    WaitAbility = 212
    Interject = 10101010

    AA = -22

    # Limit Break
    TankLB1 = 1111
    TankLB2 = 1112
    TankLB3 = 1113


# DarkKnight

class DarkKnightActions(ActionEnum):
    Unleash = 3621
    FloodOfShadow = 10
    StalwartSoul = 16468
    Quietus = 5
    AbyssalDrain = 12
    HardSlash = 3617
    SyphonStrike = 3623
    Unmend = 3624
    BloodWeapon = 3625
    Souleater = 3632
    DarkMind = 3634
    ShadowWall = 3636
    LivingDead = 3638
    SaltedEarth = 3639
    Plunge = 3640
    CarveAndSpit = 3643
    Delirium = 7390
    Bloodspiller = 7392
    TheBlackestNight = 7393
    EdgeOfShadow = 16470
    DarkMissionary = 16471
    LivingShadow = 16472
    Oblation = 25754
    SaltAndDarkness = 25755
    Shadowbringer = 25757
                             # Living Shadow actions
    PAbyssalDrain = 22
    PPlunge = 23
    PShadowbringer = 25
    PEdgeShadow = 26
    PBloodspiller = 27
    PCarveSpit = 28
    PDelay = 29

    # DOT
    SaltedEarthDOT = -14


# Gunbreaker

class GunbreakerActions(ActionEnum):
    Superbolide = 16152
    LightningShot = 16143
    RoughDivide = 16154
    NoMercy = 16138
    Bloodfest = 16164
    KeenEdge = 16137
    BrutalShell = 16139
    SolidBarrel = 16145
    BurstStrike = 16162
    GnashingFang = 16146
    SavageClaw = 16147
    WickedTalon = 16150
    JugularRip = 16156
    AbdomenTear = 16157
    EyeGouge = 16158
    Hypervelocity = 25759
    SonicBreak = 16153
    BlastingZone = 16165
    DoubleDown = 25760
    DemonSlice = 16141
    DemonSlaughter = 16149
    FatedCircle = 16163
    Nebula = 16148
    Camouflage = 16140
    HeartOfCorundum = 25758
    Aurora = 16151
    HeartOfLight = 16160
    BowShock = 16159

    # DOT
    BowShockDOT = -10
    SonicBreakDOT = -9

# Paladin

class PaladinActions(ActionEnum):
    HallowedGround = 30
    ShieldLob = 24
    Intervene = 16461
    FightOrFlight = 20
    Requiescat = 7383
    FastBlade = 9
    RiotBlade = 15
    GoringBlade = 3538
    RoyalAuthority = 3539
    Atonement = 16460
    HolySpirit = 7384
    Confiteor = 16459
    BladeOfFaith = 25748
    BladeOfTruth = 25749
    BladeOfValor = 25750
    Expiacion = 25747
    TotalEclipse = 7381
    Prominence = 16457
    CircleOfScorn = 23
    HolyCircle = 16458
    Sentinel = 17
    HolySheltron = 25746
    Intervention = 7382
    Cover = 27
    Clemency = 3541
    DivineVeil = 3540
    PassageOfArms = 7385

    # DOT
    ScornDOT = -6


# Warrior

class WarriorActions(ActionEnum):
    Holmgang = 43
    Onslaught = 7386
    Infuriate = 52
    InnerRelease = 7389
    HeavySwing = 31
    Maim = 37
    StormPath = 42
    StormEye = 45
    FellCleave = 3549
    InnerChaos = 16465
    PrimalRend = 25753
    Upheaval = 7387
    Overpower = 41
    MythrilTempest = 16462
    Decimate = 3550
    ChaoticCyclone = 16463
    Orogeny = 25752
    Vengeance = 44
    ThrillOfBattle = 40
    Equilibrium = 3552
    Bloodwhetting = 25751
    NascentFlash = 16464
    ShakeItOff = 7388
    Tomahawk = 46
    
# Melee

class MeleeActions(ActionEnum):
    SecondWind = 7541
    LegSweep = 7863
    Bloodbath = 7542
    Feint = 7549
    TrueNorth = 7546
    ArmLength = 7548
    Potion = -2
    WaitAbility = 212

    AA = -22

    # Limit Break
    MeleeLB1 = 1111
    MeleeLB2 = 1112
    MeleeLB3 = 1113

# Samurai

class SamuraiActions(ActionEnum):
    Hakaze = 7477
    Shifu = 7479
    Kasha = 7482
    Jinpu = 7478
    Gekko = 7481
    Yukikaze = 7480
    Fuko = 25780
    Oka = 7485
    Mangetsu = 7484
    Meikyo = 7499
    Higanbana = 7489
    TenkaGoken = 7488
    Midare = 7487
    OgiNamikiri = 25781
    KaeshiHiganbana = 16484
    KaeshiGoken = 16485
    KaeshiSetsugekka = 16486
    KaeshiNamikiri = 25782
    Shinten = 7490
    Kyuten = 7491
    Senei = 16481
    Guren = 7496
    Shoha = 16487
    ShohaII = 25779
    Ikishoten = 16482
    Meditate = 7497
    Hagakure = 7495
    Gyoten = 7492
    Yaten = 7493
    Enpi = 7486
    ThirdEye = 7498

    # DOT
    HiganbanaDOT = -1

    
# Reaper

class ReaperActions(ActionEnum):
    Slice = 24373
    WaxingSlice = 24374
    InfernalSlice = 24375
    ShadowOfDeath = 24378
    SoulSlice = 24380
    Gibbet = 24382
    Gallows = 24383
    SpinningScythe = 24376
    NightmareScythe = 24377
    WhorlOfDeath = 24379
    SoulScythe = 24381
    Guillotine = 24384
    BloodStalk = 24389
    UnveiledGibbet = 24390
    UnveiledGallows = 24391
    GrimSwathe = 24392
    Gluttony = 24393
    Soulsow = 24387
    HarvestMoon = 24388
    PlentifulHarvest = 24385
    Enshroud = 24394
    VoidReaping = 24395
    CrossReaping = 24396
    GrimReaping = 24397
    LemureSlice = 24399
    LemureScythe = 24400
    Communio = 24398
    ArcaneCrest = 24404
    ArcaneCircle = 24405
    HellIngress = 24402
    Regress = 24401
    Harpe = 24386

    
# Ninja

class NinjaActions(ActionEnum):
    SpinningEdge = 2240
    GustSlash = 2242
    AeolianEdge = 2255
    ArmorCrush = 3563
    ThrowingDagger = 2247
    ShadeShift = 2241
    DeathBlossom = 2254
    HakkeMujinsatsu = 16488
    Hide = 2245
    Mug = 2248
    TrickAttack = 2258
    Ten = 2259 # 18805 id from Kassatsu
    Chi = 2261 # 18806 id from Kassatsu
    Jin = 2263 # 18807 id from Kassatsu
    Ten2 = 18873 # Ten from TenChiJin
    Chi2 = 18877 # Chi from TenChiJin
    Jin2 = 18881 # Jin from TenChiJin
    FumaShuriken = 2265
    Katon = 2266
    Raiton = 2267
    Hyoton = 2268
    Huton = 2269
    Doton = 2270
    Suiton = 2271
    GokaMekkyaku = 16491
    HyoshoRanryu = 16492
    PhantomKamaitachi = 25774
    DreamWithinADream = 3566
    Huraijin = 25876
    Shukuchi = 2262
    HellfrogMedium = 7401
    Bhavacakra = 7402
    TenChiJin = 7403
    Bunshin = 16493
    FleetingRaiju = 25777
    ForkingRaiju = 25778
    Kassatsu = 2264
    Meisui = 16489

    # Shadow Actions

    BunshinPhantom = 8
    ShadowAuto = 35

    # DOT
    DotonDOT = -33


# Monk

class MonkActions(ActionEnum):
    Bootshine = 53
    TrueStrike = 54
    Demolish = 66
    DragonKick = 74
    TwinSnakes = 61
    SnapPunch = 56
    ShadowOfTheDestroyer = 25767
    FourpointFurry = 16473
    Rockbreaker = 70
    Meditation = 3546
    TheForbiddenChakra = 3547
    Enlightenment = 16474
    PerfectBalance = 69
    ElixirField = 3545
    CelestialRevolution = 25765
    RisingPhoenix = 25768
    PhantomRush = 25769
    SixSidedStar = 16476
    FormShift = 4262
    Thunderclap = 25762
    RiddleOfEarth = 7394
    RiddleOfFire = 7395
    RiddleOfWind = 25766
    Brotherhood = 7396
    Anatman = 16475
    Mantra = 65


    # DOT
    AA = -5
    DemolishDOT = -10

# Dragoon

class DragoonActions(ActionEnum):
    TrueThrust = 75
    VorpalThrust = 78
    HeavenThrust = 25771
    Disembowel = 87
    ChaoticSpring = 25772
    RaidenThrust = 16479
    FangAndClaw = 3554
    WheelingThrust = 3556
    DoomSpike = 86
    SonicThrust = 7397
    CoerthanTorment = 16477
    DraconianFury = 25770
    PiercingTalon = 90
    HighJump = 16478
    MirageDive = 7399
    SpineshatterDive = 95
    DragonFireDive = 96
    BattleLitany = 3557
    DragonSight = 7398
    LifeSurge = 83
    LanceCharge = 85
    ElusiveJump = 94
    Geirskogul = 3555
    Nastrond = 7400
    Stardiver = 16480
    WyrmwindThrust = 25773

    # DOT
    ChaoticSpringDOT = -220
    

# Physical Ranged

class RangedActions(ActionEnum):
    LegGraze = 7554
    FootGraze = 7553
    HeadGraze = 7551
    Peloton = 7557
    SecondWind = 7541
    ArmLength = 7548
    Potion = -2
    WaitAbility = 212

    AA = -23

    # Limit Break
    RangedLB1 = 1111
    RangedLB2 = 1112
    RangedLB3 = 1113


# Bard

class BardActions(ActionEnum):
    RagingStrikes = 101
    Barrage = 107
    MageBallad = 114
    ArmyPaeon = 116
    RainOfDeath = 117
    BattleVoice = 118
    EmpyrealArrow = 3558
    WanderingMinuet = 3559
    IronJaws = 3560
    WardenPaean = 3561
    Sidewinder = 3562
    PitchPerfect = 7404 # Assumed to be 3 stack
    Troubadour = 7405
    Causticbite = 7406
    Stormbite = 7407
    NatureMinne = 7408
    RefulgentArrow = 7409
    Shadowbite = 16494
    BurstShot = 16495
    ApexArow = 16496 # Assumed to be 80 gauge
    Ladonsbite = 25783
    RadiantFinale = 25785
    RepellingShot = 112
    Bloodletter = 110
    BlastArow = 25784

    # DOT
    CausticbiteDOT = -21
    StormbiteDOT = -20

# Machinist

class MachinistActions(ActionEnum):
    SplitShot = 7411
    SlugShot = 7412
    ChainSaw = 25788
    Scattergun = 25768
    Hypercharge = 17209
    Tactician = 16889
    Overdrive = 16502
    Automaton = 16501
    AirAnchor = 16500
    Bioblaster = 16499
    Drill = 16498
    AutoCrossbow = 16497
    Flamethrower = 7418 # Assumed to be for 2.5s
    BarrelStabilizer = 7414
    CleanShot = 7413
    HeatBlast = 7410
    Wildfire = 2878
    WildfireOff = -2878
    Detonator = 1111111
    GaussRound = 2874
    Ricochet = 2890
    Reassemble = 2876
    Dismantle = 2887

    # Queen actions
    Bunker = 15
    Collider = 16
    QueenAA = -24
    

# Dancer

class DancerActions(ActionEnum):
    StarfallDance = 25792
    FanDanceIV = 25791
    CuringWaltz = 16015
    Improvisation = 16014
    Flourish = 16013
    Samba = 16012
    Devilment = 16011
    EnAvant = 16010
    FanDanceIII = 16009
    FanDanceII = 16008
    FanDanceI = 16007
    ClosedPosition = 16006
    SaberDance = 16005
    StandardStep = 15997
    TechnicalStep = 15998
    Bloodshower = 15996
    RisingWindmill = 15995
    Bladeshower = 15994
    Windmill = 15993
    FountainFall = 15992
    ReverseCascade = 15991
    Fountain = 15990
    Cascade = 15989
    TechnicalFinish = 16004
    StandardFinish = 16192
    ImprovisedFinish = 25789
    Ending = 18073
    Tillana = 25790
    Emboite = 15999
    Entrechat = 16000
    Jete = 16001
    Pirouette = 16002

    AA = -22 # Dancer has Melee autos for some reason. I think that is correct.


# These are the functions we will call since they check for both cls and job_class
def name_for_id(id : int, cls : RoleEnum, job_cls : JobEnum):
    name = cls.name_for_id(id)
    if name == "Unknown": 
        name = job_cls.name_for_id(id)
    if name == "Unknown": # IF still unknown
        log_str = (
            "Unable to match ID : " + str(id) + " to an ability name in job : " + job_cls.__name__
        )
        action_logging.warning(log_str)
    return name

def id_for_name(name, cls, job_cls):
    id = cls.id_for_name(name)
    if id == -1: 
        id = job_cls.id_for_name(name)

    if id == -1 : # if still -1
        log_str = (
            "Unable to match name : " + name + " to an ability id in class : " + job_cls.__name__
        )
        
        action_logging.warning(log_str)
    return id

    


from copy import deepcopy
from ffxivcalc.Jobs.Base_Spell import DOTSpell, buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Melee.Melee_Spell import MonkSpell
Lock = 0.75 #skill animation lock - simulating 75ms ping

#Requirements

def RaptorFormRequirement(Player, Spell):
    return Player.CurrentForm == 2 or Player.FormlessFistStack > 0, -1

def CoeurlFormRequirement(Player, Spell):
    return Player.CurrentForm == 3 or Player.FormlessFistStack > 0 , -1

def ThunderclapRequirement(Player, Spell):
    return Player.ThunderclapStack > 0, Player.ThunderclapCD

def MantraRequirement(Player, Spell):
    return Player.MantraCD <= 0, Player.MantraCD

def PerfectBalanceRequirement(Player, Spell):
    return Player.PerfectBalanceStack > 0, Player.PerfectBalanceCD

def ElixirFieldRequirement(Player, Spell):
    return Player.BeastChakraType() == 1, -1

def RisingPhoenixRequirement(Player, Spell):
    return Player.BeastChakraType() == 3, -1

def CelestialRevolutionRequirement(Player, Spell):
    return Player.BeastChakraType() == 2, -1

def NadiRequirement(Player, Spell):
    return Player.MasterGauge[0] and Player.MasterGauge[4], -1

def BrotherhoodRequirement(Player, Spell):
    return Player.BrotherhoodCD <= 0, Player.BrotherhoodCD

def RiddleOfEarthRequirement(Player, Spell):
    return Player.RiddleOfEarthStack > 0, Player.RiddleOfEarthCD

def RiddleOfFireRequirement(Player, Spell):
    return Player.RiddleOfFireCD <= 0, Player.RiddleOfFireCD

def RiddleOfWindRequirement(Player, Spell):
    return Player.RiddleOfWindCD <= 0, Player.RiddleOfWindCD

def ChakraRequirement(Player, Spell):
    return Player.MaxChakraGate == 5, -1

#Apply

def ApplyFormShift(Player, Enemy):
    Player.FormlessFistStack += 1

def ApplyChakra(Player, Enemy):
    Player.UsedChakraGate += 5
    Player.MaxChakraGate = 0

def ApplyRiddleOfWind(Player, Enemy):
    Player.RiddleOfWindCD = 90
    Player.RiddleOfWindTimer = 15
                             # Delay reduced by 50%.
                             # Since it only says it reduces the delay by 50%
                             # This does not affect the current DOTTimer of the auto.
                             # This could be a wrong interpretation.
                             # This will only update the Player.currentDelay
    Player.autoHaste = 50
    Player.hasteHasChanged = True
    Player.EffectCDList.append(RiddleOfWindCheck)

def ApplyRiddleOfFire(Player, Enemy):
    Player.RiddleOfFireCD = 60
    Player.RiddleOfFireTimer = 20
    Player.EffectCDList.append(RiddleOfFireCheck)
    Player.buffList.append(RiddleOfFireBuff)

                                     # Only doing this if SavePreBakedAction is true and if the PreBaked player is the monk
    #if Player.CurrentFight.SavePreBakedAction and Player == Player.CurrentFight.PlayerList[Player.CurrentFight.PlayerIDSavePreBakedAction]:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , RiddleOfFireBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyRiddleOfEarth(Player, Enemy):
    Player.RiddleOfEarthCD = 120

def ApplyElixirField(Player, Enemy):
    Player.ResetMasterGauge()

    Player.MasterGauge[0] = True #Opening Lunar Nadi
    Player.FormlessFistStack += 1

def ApplyCelestialRevolution(Player, Enemy):
    Player.ResetMasterGauge()
    Player.FormlessFistStack += 1

    if Player.MasterGauge[0]: Player.MasterGauge[-1] = True #If lunar already opened open Solar
    else: Player.MasterGauge[0] = True

def ApplyRisingPhoenix(Player, Enemy):
    Player.ResetMasterGauge()

    Player.MasterGauge[-1] = True #Opening Solar Nadi
    Player.FormlessFistStack += 1

def ApplyPhantomRush(Player, Enemy):
    Player.ResetMasterGauge()
    Player.MasterGauge[0], Player.MasterGauge[-1] = False, False
    #Reset the whole Gauge
    Player.FormlessFistStack += 1

def ApplyThunderclap(Player, Enemy):
    if Player.ThunderclapStack == 3:
        Player.EffectCDList.append(ThunderclapStackCheck)
        Player.ThunderclapCD = 30
    Player.ThunderclapStack -= 1


def ApplyOpoOpo(Player, Enemy):
    FormlessStackCheck(Player)
    Player.CurrentForm = 1

def ApplyRaptor(Player, Enemy):
    FormlessStackCheck(Player)
    Player.CurrentForm = 2 #Into Raptor

def ApplyCoeurl(Player, Enemy):
    FormlessStackCheck(Player)
    Player.CurrentForm = 3 #Into Coeurl

def ApplyMantra(Player, Enemy):
    Player.MantraCD = 90

def ApplyPerfectBalance(Player, Enemy):
    if Player.PerfectBalanceStack == 2:
        Player.EffectCDList.append(PerfectBalanceStackCheck)
        Player.PerfectBalanceCD = 40
    Player.PerfectBalanceStack -= 1
    Player.PerfectBalanceEffectStack = 3
    Player.FormlessFistStack += 3
    Player.EffectList.append(PerfectBalanceEffect)

def ApplyMeditation(Player, Enemy):
    if not Player.CurrentFight.FightStart:
        Player.MaxChakraGate = 5
        Player.ExpectedChakraGate = 5
    else:
        Player.OpenChakra()

def ApplyBrotherhood(Player, Enemy):
    # We will give everyone the 5% buff, and the effect that will
    # check for weaponskil/casted actions and we will assume that
    # all of them open a chakra gate

    def MeditativeBrotherhoodEffect(Target, Spell):
        if Spell.GCD:
            Player.OpenChakra() #Adds 1 chakra
            Player.ExpectedChakraGate += 0.2 #Add expected Chakra
    
    def MeditativeBrotherhoodCheck(Target, Enemy):
        if Target.MeditativeBrotherhoodTimer <= 0:
            Target.EffectList.remove(MeditativeBrotherhoodEffect)
            Target.EffectToRemove.append(MeditativeBrotherhoodCheck)
            Target.buffList.remove(BrotherhoodBuff)


    def MonkMeditativeEffect(Target, Spell):
        #This one goes on the Monk
        if Spell.GCD:
            Target.ExpectedChakraGate += 1
            Target.OpenChakra()
            
    def MonkMeditativeCheck(Target, Spell):
        if Target.MeditativeBrotherhoodTimer <= 0:
            Target.EffectList.remove(MonkMeditativeEffect)
            Target.EffectToRemove.append(MonkMeditativeCheck)
            Target.buffList.remove(BrotherhoodBuff)



    for Target in Player.CurrentFight.PlayerList:
        if Target == Player:
            #Need special effect for the Monk
            Target.EffectList.append(MonkMeditativeEffect)
            Target.EffectCDList.append(MonkMeditativeCheck)
            Target.buffList.append(BrotherhoodBuff)
            Target.MeditativeBrotherhoodTimer = 15 #Adding buff
        else:
            #Applying MeditativeBrotherhood to every other player
            Target.EffectList.append(MeditativeBrotherhoodEffect)
            Target.EffectCDList.append(MeditativeBrotherhoodCheck)
            Target.buffList.append(BrotherhoodBuff)
            Target.MeditativeBrotherhoodTimer = 15 #Adding buff

                                 # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15 , BrotherhoodBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)
#Effect

def LeadenFistEffect(Player, Spell):
    if Spell.id == Bootshine.id: Spell.Potency += 100

#Check

def RiddleOfWindCheck(Player, Enemy):
    if Player.RiddleOfWindTimer <= 0:
        Player.autoHaste = 0
        Player.hasteHasChanged = True
        Player.EffectToRemove.append(RiddleOfWindCheck)

def RiddleOfFireCheck(Player, Enemy):
    if Player.RiddleOfFireTimer <= 0:
        Player.buffList.remove(RiddleOfFireBuff)
        Player.EffectToRemove.append(RiddleOfFireCheck)

def RiddleOfEarthStackCheck(Player, Enemy):
    if Player.RiddleOfEarthCD <= 0:
        if Player.RiddleOfEarthStack == 2:
            Player.EffectToRemove.append(RiddleOfEarthStackCheck)
        else:
            Player.RiddleOfEarthCD = 30
        Player.RiddleOfEarthStack += 1

def PerfectBalanceStackCheck(Player, Enemy):
    if Player.PerfectBalanceCD <= 0:
        if Player.PerfectBalanceStack == 1:
            Player.EffectToRemove.append(PerfectBalanceStackCheck)
        else:
            Player.PerfectBalanceTimer = 40
        Player.PerfectBalanceStack += 1

def ThunderclapStackCheck(Player, Enemy):
    if Player.ThunderclapCD <= 0:
        if Player.ThunderclapStack == 2:
            Player.EffectToRemove.append(ThunderclapStackCheck)
        else:
            Player.ThunderclapTimer = 30
        Player.ThunderclapStack += 1

def DemolishDOTCheck(Player, Enemy):
    if Player.DemolishDOTTimer <= 0:
        Player.DOTList.remove(Player.DemolishDOT)
        Player.DemolishDOT = None
        Player.EffectToRemove.append(DemolishDOTCheck)

def DisciplinedFistCheck(Player, Enemy):
    if Player.DisciplinedFistTimer <= 0:
        Player.buffList.remove(DisciplinedFistBuff)
        Player.EffectToRemove.append(DisciplinedFistCheck)

def LeadenFistCheck(Player, Enemy):
    if Player.LeadenFistTimer <= 0:
        Player.EffectList.remove(LeadenFistEffect)
        Player.EffectToRemove.append(LeadenFistCheck)

#Combo Effect

def PerfectBalanceEffect(Player, Spell):
    if Spell.GCD:
        if Spell.id == Bootshine.id or Spell.id == ShadowOfTheDestroyer.id or Spell.id == DragonKick.id:
            Player.addBeastChakra(1)
            Player.PerfectBalanceEffectStack -= 1
        elif Spell.id == TrueStrike.id or Spell.id == TwinSnakes.id or Spell.id == FourpointFurry.id:
            Player.addBeastChakra(2)
            Player.PerfectBalanceEffectStack -= 1
        elif Spell.id == Demolish.id or Spell.id == SnapPunch.id or Spell.id == Rockbreaker.id:
            Player.addBeastChakra(3)
            Player.PerfectBalanceEffectStack -= 1

    if Player.PerfectBalanceEffectStack == 0:
        Player.EffectToRemove.append(PerfectBalanceEffect)

def ComboEffect(Player, Spell):
    # This effect will always be on and will check what form the player is in and use the
    # necessary effect
    if Spell.GCD:
        if Player.FormlessFistStack > 0 :#Then we are in formless, so we can do whichever

            Player.UsedFormlessStack = True

            #Opo-Opo Form
            if Spell.id == Bootshine.id or Spell.id == ShadowOfTheDestroyer.id:
                Player.GuaranteedCrit = True
            elif Spell.id == DragonKick.id:
                #Grants Leaden Fist
                if Player.LeadenFistTimer == 0: #If not already applied
                    Player.EffectList.append(LeadenFistEffect)
                    Player.EffectCDList.append(LeadenFistCheck)
                Player.LeadenFistTimer = 30

            #Raptor Form

            elif Spell.id == TrueStrike.id: #Nothing happens
                pass
            elif Spell.id == TwinSnakes.id or Spell.id == FourpointFurry.id:
                if Player.DisciplinedFistTimer == 0:
                    Player.buffList.append(DisciplinedFistBuff)
                    Player.EffectCDList.append(DisciplinedFistCheck)
                Player.DisciplinedFistTimer = 15

            #Coeurl Form

            elif Spell.id == Demolish.id:
                if Player.DemolishDOT == None:
                    Player.DemolishDOT = deepcopy(DemolishDOT)
                    Player.EffectCDList.append(DemolishDOTCheck)
                    Player.DOTList.append(Player.DemolishDOT)
                else: Player.DemolishDOT.resetBuffSnapshot() # If already applied reset snapshot
                Player.DemolishDOTTimer = 18
            elif Spell.id == SnapPunch.id or Spell.id == Rockbreaker.id:
                pass #Nothing happens

        #Not formless
        elif Player.CurrentForm == 1: #Opo-Opo form
            if Spell.id == Bootshine.id or Spell.id == ShadowOfTheDestroyer.id:
                Player.GuaranteedCrit = True
            elif Spell.id == DragonKick.id:
                #Grants Leaden Fist
                if Player.LeadenFistTimer == 0: #If not already applied
                    Player.EffectList.append(LeadenFistEffect)
                    Player.EffectCDList.append(LeadenFistCheck)
                Player.LeadenFistTimer = 30
        elif Player.CurrentForm == 2 : #Raptor Form
            if Spell.id == TrueStrike.id: #Nothing happens
                pass
            elif Spell.id == TwinSnakes.id or Spell.id == FourpointFurry.id:
                if Player.DisciplinedFistTimer == 0:
                    Player.buffList.append(DisciplinedFistBuff)
                    Player.EffectCDList.append(DisciplinedFistCheck)
                Player.DisciplinedFistTimer = 15
        elif Player.CurrentForm == 3 : #Coeurl Form
            if Spell.id == Demolish.id:
                if Player.DemolishDOT == None:
                    Player.DemolishDOT = deepcopy(DemolishDOT)
                    Player.EffectCDList.append(DemolishDOTCheck)
                    Player.DOTList.append(Player.DemolishDOT)
                else: Player.DemolishDOT.resetBuffSnapshot() # If already applied reset snapshot
                Player.DemolishDOTTimer = 18
            elif Spell.id == SnapPunch.id or Spell.id == Rockbreaker.id:
                pass #Nothing happens


#Other

def FormlessStackCheck(Player):
    if Player.UsedFormlessStack:
        Player.UsedFormlessStack = False
        Player.FormlessFistStack -= 1

#Opo-opo form  -> Raptor Form
Bootshine = MonkSpell(53, True, 2.5, 210, ApplyRaptor, [], True, False, type = 2)
DragonKick = MonkSpell(74, True, 2.5, 320, ApplyRaptor, [], True, False, type = 2)
ShadowOfTheDestroyer = MonkSpell(25767, True,2, 110, ApplyRaptor, [], True, False, type = 2) #AOE of Bootshine

#Raptor form combo -> Coeurl form
TrueStrike = MonkSpell(54, True, 2.5, 300, ApplyCoeurl, [RaptorFormRequirement], True, False, type = 2)
TwinSnakes = MonkSpell(61, True, 2.5, 280, ApplyCoeurl, [RaptorFormRequirement], True, False, type = 2)
FourpointFurry = MonkSpell(16473, True, 2.5, 120, ApplyCoeurl, [RaptorFormRequirement], True, False, type = 2) #AOE of Twinsnakes

#Coeurl form combo -> Opo-opo form
Demolish = MonkSpell(66, True, 2.5, 130, ApplyOpoOpo, [CoeurlFormRequirement], True, False, type = 2)
DemolishDOT = DOTSpell(-10, 70, True)
SnapPunch = MonkSpell(56, True,2.5 ,310, ApplyOpoOpo, [CoeurlFormRequirement], True, False, type = 2)
Rockbreaker = MonkSpell(70, True, 2.5, 130, ApplyOpoOpo, [CoeurlFormRequirement], True, False, type = 2)

#Chakra
TheForbiddenChakra = MonkSpell(3547, False, 0, 340, ApplyChakra, [ChakraRequirement], False, False )
Enlightenment = MonkSpell(16474, False, 0, 170, ApplyChakra, [ChakraRequirement], False, False)
#Masterful Blitz
ElixirField = MonkSpell(3545, True, 2.5, 600, ApplyElixirField, [ElixirFieldRequirement], True, False, type = 2)
CelestialRevolution = MonkSpell(25765, True, 2.5, 450, ApplyCelestialRevolution, [CelestialRevolutionRequirement], True, False, type = 2)
RisingPhoenix = MonkSpell(25768, True, 2.5, 700, ApplyRisingPhoenix, [RisingPhoenixRequirement], True, False, type = 2)
PhantomRush = MonkSpell(25769, True, 2.5, 1150, ApplyPhantomRush, [RisingPhoenixRequirement, NadiRequirement], True, False, type = 2)
#Other GCD
Meditation = MonkSpell(3546, True, 1, 0, ApplyMeditation, [], False, False, type = 2)
FormShift = MonkSpell(4262, True, 2.5, 0, ApplyFormShift, [], False, False, type = 2)
SixSidedStar = MonkSpell(16476, True, 4, 550, empty, [], True, False, type = 2)
#oGCD
PerfectBalance = MonkSpell(69, False, 0, 0, ApplyPerfectBalance, [PerfectBalanceRequirement], False, False)
Brotherhood = MonkSpell(7396, False, 0, 0, ApplyBrotherhood, [BrotherhoodRequirement], False, False)
RiddleOfFire = MonkSpell(7395, False, 0, 0, ApplyRiddleOfFire, [RiddleOfFireRequirement], False, False)
RiddleOfEarth = MonkSpell(7394, False, 0, 0, ApplyRiddleOfEarth, [RiddleOfEarthRequirement], False, False)
RiddleOfWind = MonkSpell(25766, False, 0, 0, ApplyRiddleOfWind, [RiddleOfWindRequirement], False, False)
#non Damage oGCD
Anatman = MonkSpell(16475, False, 0, 0, empty, [], False, False)
Thunderclap = MonkSpell(25762, False, 0, 0, ApplyThunderclap, [ThunderclapRequirement],False, False)
Mantra = MonkSpell(65, False, 0, 0, ApplyMantra, [MantraRequirement], False, False)
#Buff
DisciplinedFistBuff = buff(1.15,name="Disciplined Fist")
BrotherhoodBuff = buff(1.05,name="Bro")
RiddleOfFireBuff = buff(1.15,name="Riddle Of Fire")



MonkAbility = {
53 : Bootshine,
54 : TrueStrike,
66 : Demolish,
74 : DragonKick,
61 : TwinSnakes,
56 : SnapPunch,
25767 : ShadowOfTheDestroyer,
16473 : FourpointFurry,
70 : Rockbreaker,
3546 : Meditation,
3547 : TheForbiddenChakra,
16474 : Enlightenment,
69 : PerfectBalance,
3545 : ElixirField,
25765 : CelestialRevolution,
25768 : RisingPhoenix,
25769 : PhantomRush,
16476 : SixSidedStar,
4262 : FormShift,
25762 : Thunderclap,
7394 : RiddleOfEarth,
7395 : RiddleOfFire,
25766 : RiddleOfWind,
7396 : Brotherhood,
16475 : Anatman,
65 : Mantra

}


from ffxivcalc.Jobs.Base_Spell import DOTSpell, buff, empty, buffHistory, buffPercentHistory
from ffxivcalc.Jobs.Melee.Melee_Spell import DragoonSpell
import copy

#Requirement

def TrueThrustRequirement(Player, Spell):
    return not Player.DraconianFire, -1

def WheelingThrustRequirement(Player, Spell):
    return Player.WheelInMotion, -1

def FangAndClawRequirement(Player, Spell):
    return Player.FangAndClaw, -1

def LanceChargeRequirement(Player, Spell):
    return Player.LanceChargeCD <= 0, Player.LanceChargeCD

def BattleLitanyRequirement(Player, Spell):
    return Player.BattleLitanyCD <= 0, Player.BattleLitanyCD

def DragonSightRequirement(Player, Spell):
    return Player.DragonSightCD <= 0, Player.DragonSightCD

def GeirskogulRequirement(Player, Spell):
    return Player.GeirskogulCD <= 0 and not Player.LifeOfTheDragon, Player.GeirskogulCD

def NastrondRequirement(Player, Spell):
    return Player.LifeOfTheDragon and Player.NastrondCD <= 0, Player.NastrondCD

def HighJumpRequirement(Player, Spell):
    return Player.HighJumpCD <= 0, Player.HighJumpCD

def MirageDiveRequirement(Player, Spell):
    return Player.DiveReady, -1

def SpineshafterRequirement(Player, Spell):
    return Player.SpineshafterStack > 0, Player.SpineshafterCD

def LifeSurgeRequirement(Player, Spell):
    return Player.LifeSurgeStack > 0, Player.LifeSurgeCD

def StardiverRequirement(Player, Spell):
    return Player.StardiverCD <= 0 and Player.LifeOfTheDragon, Player.StardiverCD

def RaidenThrustRequirement(Player, Spell):
    return Player.DraconianFire, -1

def WyrmwindThrustRequirement(Player, Spell):
    return Player.FirstmindGauge == 2 and Player.WyrmwindThrustCD <= 0, -1

def DragonFireDiveRequirement(Player, Spell):
    return Player.DragonFireDiveCD <= 0, Player.DragonFireDiveCD

#Apply

def ApplyDoomSpike(Player, Enemy):
    if not (DoomSpikeCombo in Player.EffectList ): Player.EffectList.append(DoomSpikeCombo)

def ApplyDragonFireDive(Player, Enemy):
    Player.DragonFireDiveCD = 120

def ApplyTrueThrust(Player, Enemy):
    if not (TrueThrustCombo in Player.EffectList) : Player.EffectList.append(TrueThrustCombo)

def ApplyWheelingThrust(Player, Enemy):
    Player.WheelInMotion = False
    
    if not Player.LanceMastery: 
        Player.FangAndClaw = True
        Player.LanceMastery = True
        Player.EffectList.append(LanceMasteryCombo)
    else: 
        Player.LanceMastery = False


def ApplyFangAndClaw(Player, Enemy):
    Player.FangAndClaw = False

    if not Player.LanceMastery: 
        Player.WheelInMotion = True
        Player.LanceMastery = True
        Player.EffectList.append(LanceMasteryCombo)
    else:
        Player.LanceMastery = False
def ApplyLanceCharge(Player, Enemy):
    Player.LanceChargeCD = 60
    Player.LanceChargeTimer = 20
    Player.EffectCDList.append(LanceChargeCheck)
    Player.buffList.append(LanceChargeBuff)

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , LanceChargeBuff.MultDPS)
    #    Player.PercentBuffHistory.append(history)

def ApplyBattleLitany(Player, Enemy):
    Player.BattleLitanyCD = 120
    Player.BattleLitanyTimer = 15

    #Will give each person in the fight the buff
    for player in Player.CurrentFight.PlayerList:  
        player.CritRateBonus += 0.1
    Player.EffectCDList.append(BattleLitanyCheck)

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffHistory(fight.TimeStamp, fight.TimeStamp + 15)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].BattleLitanyHistory.append(history)

def ApplyGeirskogul(Player, Enemy):
    Player.GeirskogulCD = 30

    if Player.DragonGauge == 2 : 
        Player.LifeOfTheDragon = True
        Player.LifeOfTheDragonTimer = 30
        Player.EffectCDList.append(LifeOfTheDragonCheck)
        Player.DragonGauge = 0

def ApplyHighJump(Player, Enemy):
    Player.HighJumpCD = 30
    Player.DiveReady = True

def ApplyMirageDive(Player, Enemy):
    Player.DiveReady = False
    Player.DragonGauge = min(2, Player.DragonGauge + 1)

def ApplySpineshafter(Player, Enemy):
    if Player.SpineshafterStack == 2:
        Player.EffectCDList.append(SpineshafterStackCheck)
        Player.SpineshafterCD = 60
    Player.SpineshafterStack -= 1

def ApplyLifeSurge(Player, Enemy):
    if Player.LifeSurgeStack == 2:
        Player.EffectCDList.append(LifeSurgeStackCheck)
        Player.LifeSurgeCD = 40
    Player.LifeSurgeStack -= 1

    Player.NextCrit = True

def ApplyWyrmwindThrust(Player, Enemy):
    Player.FirstmindGauge = 0
    Player.WyrmwindThrustCD = 10

def ApplyNastrond(Player, Enemy):
    Player.NastrondCD = 10

def ApplyStardiver(Player, Enemy):
    Player.StardiverCD = 30

def ApplyRaidenThrust(Player, Enemy):
    Player.FirstmindGauge = min(2, Player.FirstmindGauge + 1)
    Player.DraconianFire = False

    ApplyTrueThrust(Player, Enemy) #Since considered as first of combo

#Effect

def DoomSpikeCombo(Player, Spell):
    if Spell.id == SonicThrust.id:
        Spell.Potency += 20
            #Gain PowerSurge, 10% damage

        if Player.PowerSurgeTimer <= 0: #Not already applied
            Player.buffList.append(PowerSurgeBuff)
            Player.EffectCDList.append(PowerSurgeCheck)
        Player.PowerSurgeTimer = 30
        Player.EffectList.append(SonicThrustCombo)
        Player.EffectToRemove.append(DoomSpikeCombo)

def SonicThrustCombo(Player, Spell):
    if Spell.id == CoerthanTorment.id:
        Spell.Potency += 50
        Player.DraconianFire = True
        Player.EffectToRemove.append(SonicThrustCombo)



def TrueThrustCombo(Player, Spell):
    if Spell.id == Disembowel.id:
        Spell.Potency += 110

        #Gain PowerSurge, 10% damage

        if Player.PowerSurgeTimer <= 0: #Not already applied
            Player.buffList.append(PowerSurgeBuff)
            Player.EffectCDList.append(PowerSurgeCheck)
        Player.PowerSurgeTimer = 30

        Player.EffectList.append(DisembowelCombo)
        Player.EffectToRemove.append(TrueThrustCombo)

    elif Spell.id == VorpalThrust.id:
        Spell.Potency += 150
        Player.EffectList.append(VorpalThrustCombo)
        Player.EffectToRemove.append(TrueThrustCombo)

def VorpalThrustCombo(Player, Spell):
    if Spell.id == HeavenThrust.id:
        Spell.Potency += 380
        Player.FangAndClaw = True
        Player.EffectToRemove.append(VorpalThrustCombo)

def DisembowelCombo(Player, Spell):
    if Spell.id == ChaoticSpring.id:
        Spell.Potency +=160
        if Player.ChaoticSpringDOT == None:
            Player.ChaoticSpringDOT = copy.deepcopy(ChaoticSpringDOT)
            Player.DOTList.append(Player.ChaoticSpringDOT)
            Player.EffectCDList.append(ChaoticSpringDOTCheck)
        else : Player.ChaoticSpringDOT.resetBuffSnapshot() # If already applied reset snapshot
        Player.ChaoticSpringDOTTimer = 24

        Player.WheelInMotion = True 
        Player.EffectToRemove.append(DisembowelCombo)

def LanceMasteryCombo(Player, Spell):
    if Spell.id == FangAndClaw.id or Spell.id == WheelingThrust.id:
        Spell.Potency += 100
        Player.EffectToRemove.append(LanceMasteryCombo)
        Player.DraconianFire = True
    elif isinstance(Spell, DragoonSpell) and Spell.Weaponskill:
        Player.FangAndClaw = False
        Player.WheelInMotion = False
        Player.LanceMastery = False
        Player.EffectToRemove.append(LanceMasteryCombo)



#Check

def ChaoticSpringDOTCheck(Player, Enemy):
    if Player.ChaoticSpringDOTTimer <= 0:
        Player.DOTList.remove(Player.ChaoticSpringDOT)
        Player.ChaoticSpringDOT = None
        Player.EffectToRemove.append(ChaoticSpringDOTCheck)


def LifeOfTheDragonCheck(Player, Enemy):
    if Player.LifeOfTheDragonTimer <= 0:
        Player.EffectToRemove.append(LifeOfTheDragonCheck)
        Player.LifeOfTheDragon = False

def LifeSurgeStackCheck(Player, Enemy):
    if Player.LifeSurgeCD <= 0:
        if Player.LifeSurgeStack == 1:
            Player.EffectToRemove.append(LifeSurgeStackCheck)
        else:
            Player.LifeSurgeCD = 45
        Player.LifeSurgeStack += 1

def SpineshafterStackCheck(Player, Enemy):
    if Player.SpineshafterCD <= 0:
        if Player.SpineshafterStack == 1:
            Player.EffectToRemove.append(SpineshafterStackCheck)
        else:
            Player.SpineShifterCD = 60
        Player.SpineshafterStack += 1

def BattleLitanyCheck(Player, Enemy):
    if Player.BattleLitanyTimer <= 0:
        for player in Player.CurrentFight.PlayerList:  player.CritRateBonus -= 0.1 #Removing buff
        Player.EffectToRemove.append(BattleLitanyCheck)

def LanceChargeCheck(Player, Enemy):
    if Player.LanceChargeTimer <= 0:
        #input("Removing lance charge")
        Player.buffList.remove(LanceChargeBuff)
        Player.EffectToRemove.append(LanceChargeCheck)

def PowerSurgeCheck(Player, Enemy):
    #input("in check : " + str(Player.PowerSurgeTimer))
    if Player.PowerSurgeTimer <= 0:
        #input("Removing powersurge")
        Player.EffectToRemove.append(PowerSurgeCheck)
        Player.buffList.remove(PowerSurgeBuff)

#GCD
#Combo Action
TrueThrust = DragoonSpell(75, True, 2.5, 230, ApplyTrueThrust, [TrueThrustRequirement], True, type = 2)
Disembowel = DragoonSpell(87, True, 2.5, 140, empty, [], True, type = 2)
VorpalThrust = DragoonSpell(78, True, 2.5, 130, empty, [], True, type = 2)
ChaoticSpring = DragoonSpell(25772, True, 2.5, 140, empty, [], True, type = 2)
ChaoticSpringDOT = DOTSpell(-220, 45, True)
HeavenThrust = DragoonSpell(25771, True, 2.5, 100, empty, [], True, type = 2)

WheelingThrust = DragoonSpell(3556, True, 2.5, 300, ApplyWheelingThrust, [WheelingThrustRequirement], True , type = 2)
FangAndClaw = DragoonSpell(3554, True, 2.5, 300, ApplyFangAndClaw, [FangAndClawRequirement], True, type = 2)
RaidenThrust = DragoonSpell(16479, True, 2.5, 280, ApplyRaidenThrust, [RaidenThrustRequirement], True, type = 2)
PiercingTalon = DragoonSpell(90, True, 2.5, 150, empty, [], True, type = 2)

#AOE combo Action
DoomSpike = DragoonSpell(86, True, 2.5, 110, ApplyDoomSpike, [], True, type = 2)
SonicThrust = DragoonSpell(7397, True, 2.5, 100, empty, [], True, type = 2)
CoerthanTorment = DragoonSpell(16477, True, 2.5, 100, empty, [], True, type = 2)
DraconianFury = DragoonSpell(25770, True, 2.5, 130, ApplyRaidenThrust, [RaidenThrustRequirement], True, type = 2) #AOE version of Raiden Thrust

#oGCD
LanceCharge = DragoonSpell(85, False, 0, 0, ApplyLanceCharge, [LanceChargeRequirement], False)
BattleLitany = DragoonSpell(3557, False, 0, 0, ApplyBattleLitany, [BattleLitanyRequirement], False)
Geirskogul = DragoonSpell(3555, False, 0, 260, ApplyGeirskogul, [GeirskogulRequirement], False)
Nastrond = DragoonSpell(7400, False, 0, 300, ApplyNastrond, [NastrondRequirement], False)
HighJump = DragoonSpell(16478, False, 0, 400, ApplyHighJump, [HighJumpRequirement], False)
MirageDive = DragoonSpell(7399, False, 0, 200, ApplyMirageDive, [MirageDiveRequirement], False)
SpineshafterDive = DragoonSpell(95, False, 0, 250, ApplySpineshafter, [SpineshafterRequirement], False)
LifeSurge = DragoonSpell(83, False, 0, 0, ApplyLifeSurge, [LifeSurgeRequirement], False)
Stardiver = DragoonSpell(16480, False, 0, 620, ApplyStardiver, [StardiverRequirement], False)
WyrmwindThrust = DragoonSpell(25773, False, 0, 420, ApplyWyrmwindThrust, [WyrmwindThrustRequirement], False)
DragonFireDive = DragoonSpell(96, False, 0, 300, ApplyDragonFireDive, [DragonFireDiveRequirement], False)

#funny oGCD
ElusiveJump = DragoonSpell(94, False, 0, 0, empty, [], 0) #No requirement, I will let a Dragoon spam this shit


#buff
RightEyeBuff = buff(1.1,name="RE")
LeftEyeBuff = buff(1.05,name="LE")
LanceChargeBuff = buff(1.1,name="Lance Charge")
PowerSurgeBuff = buff(1.1,name="Power Surge")

def DragonSight(Target):

    def DragonSightCheck(Player, Enemy):
        if Player.DragonSightTimer <= 0:
            Player.buffList.remove(RightEyeBuff)
            Target.buffList.remove(LeftEyeBuff)
            Player.EffectToRemove.append(DragonSightCheck)

    def ApplyDragonSight(Player, Enemy):
                             # Refreshing target playerID
        DragonSightSpell.TargetID = Target.playerID

        Player.DragonSightCD = 120
        Player.DragonSightTimer = 20

        Player.buffList.append(RightEyeBuff)
        Target.buffList.append(LeftEyeBuff)
        Player.EffectCDList.append(DragonSightCheck)

                                         # Only doing this if SavePreBakedAction is true
        #if Player.CurrentFight.SavePreBakedAction:
        #    fight = Player.CurrentFight
        #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , LeftEyeBuff.MultDPS)
        #    drgHistory = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20, RightEyeBuff.MultDPS) # Adding to DRG instead of chcking
        #    Target.PercentBuffHistory.append(history)
        #    Player.PercentBuffHistory.append(drgHistory)

    DragonSightSpell = DragoonSpell(7398, False, 0, 0, ApplyDragonSight, [DragonSightRequirement], False)
    DragonSightSpell.TargetID = Target.playerID
    DragonSightSpell.TargetPlayerObject = Target
    return DragonSightSpell


DragoonAbility = {
75 : TrueThrust,
78 : VorpalThrust,
25771 : HeavenThrust,
87 : Disembowel,
25772 : ChaoticSpring,
16479 : RaidenThrust,
3554 : FangAndClaw,
3556 : WheelingThrust,
86 : DoomSpike,
7397 : SonicThrust,
16477 : CoerthanTorment,
25770 : DraconianFury,
90 : PiercingTalon,
16478 : HighJump,
7399 : MirageDive,
95 : SpineshafterDive,
96 : DragonFireDive,
3557 : BattleLitany,
7398 : DragonSight,
83 : LifeSurge,
85 : LanceCharge,
94 : ElusiveJump,
3555 : Geirskogul,
7400 : Nastrond,
16480 : Stardiver,
25773 : WyrmwindThrust
}


import os
from ffxivcalc.UI.SimulationInput import ExecuteMemoryCode
from ffxivcalc.helperCode.helper_backend import AskInput, ImportFightBackend, MergeFightBackEnd, SaveFight, SimulateFightBackend,GenerateLayoutBackend
from ffxivcalc.Jobs.PlayerEnum import JobEnum

def SimulateFightMemory():
    os.system('CLS') #clearing HUD
    #This will simulate the fight in memory

    #We will first go through some options
    print(
        "===================== SIMULATOR ====================="
    )

    #Looking at all fights we have saved

    cur_dir = os.getcwd()

    #the saved directory should be in that same folder

    saved_dir = cur_dir + "\\saved"

    saved_fight = os.listdir(saved_dir)
    number_saved_fight = len(saved_fight)
    selected_fight = ""

    if number_saved_fight == 0: #No fight saved
        print("No saved fights were found. Hit any button to return to the main menu.")
        input("...")
        return
    else: #Saved fight
        print(
            "Select one of the saved fights : "+ "\n" + 
            "=========================================================="
            )
        for i in range(1,number_saved_fight+1):
            print(str(i) + " -> " + saved_fight[i-1])

        print("==========================================================")

        userInput = int(AskInput(number_saved_fight))

        selected_fight = saved_fight[userInput-1] #Selecting fight

    
    SimulateFightBackend(saved_dir + "\\" + selected_fight)

    

def ImportFight():
    os.system('CLS') #clearing HUD
    print(
        "===================== IMPORT FIGHT FROM FFLOGS =====================" + "\n" +
        "THIS IS VERY EXPERIMENTAL" + "\n" +
        "The program does not check if the fight is valid, and will simply result in a crash if it is not."
    )

    fightID = input("Please enter the fightID : ")
    fightNumber = input("Please enter the fight's number : ")

    print("Converting FFLogs fight into Event object...")

    Event = ImportFightBackend(fightID, fightNumber)

    userInput = input("What name do you wish to save this fight as : ")

    SaveFight(Event, 0, 1000, userInput)




def CreateFight():
    os.system('CLS') #clearing HUD

def Credits():
    os.system('CLS') #clearing HUD
    #This menu displays credits
    print(
        "=================================================== CREDITS =======================================================" + "\n" + 
        "This program was made mainly by myself, but I was helped by a lot of friends" + "\n" +
        "and other people without whom this project would not have been possible." + "\n" +
        "I listed here all those that helped with a short description of what they did." + "\n" +
        "I personally thank everyone here for the help they brought :)" + "\n" +
        "I will put their discord IDs in case you wish to contact them." + "\n" +
        "=================================================================================================================" + "\n" +
        "Pythagoras#6312 -> That's me! Did most of it" + "\n" +
        "ne0dym#7837 -> DarkKnight code and other revisions" + "\n" +
        "Bri-kun#6539 -> Networking code and other FFLogsAPI related stuff" + "\n" +
        "javaJake#0001 -> Theoretical damage computation" + "\n" +
        "Saint Germain#0280 -> Bit of Monk code" + "\n" +
        "apollo#3810 -> Put me in contact with javajake and also provided some background code for working with FFLogsAPI" + "\n" +
        "Ikuni#2735 -> Helped by verifying and correcting some scalar values in the damage formulas" + "\n" +
        "redshadowhero#8631 -> Helped restructure the code in a better way and gave other advices" + "\n" +
        "\n" +
        "I also give my thanks to The Balance, which provided me with a lot of knowleadge about the different" + "\n" +
        "jobs so I could implement them and also gave me access to resources like the Blackmage gear comparison" + "\n" +
        "spreadsheet which I used a lot when trying to better understand how damage computation work in this game." + "\n" +
        "================================================================================================================="
    )

    input("Press Enter to go back to the Main Menu : ")


def GenerateLayout():
    os.system('CLS') #clearing HUD
    print(
        "===================== GENERATING LAYOUT =====================" + "\n" 
    )

    namefile = input("What name do you wish to save this fight as : ")

    player_list = []

    while True:
        os.system('CLS') #clearing HUD
        print(
        "===================================================================" + "\n" + 
        "Select the job you want to add or enter 20" + "\n" + 
        "===================================================================" + "\n" + 
        "1- BlackMage" + "\n" + 
        "2- RedMage"  + "\n" + 
        "3- Summoner" + "\n" + 
        "4- Ninja" + "\n" + 
        "5- Samurai" + "\n" + 
        "6- Monk" + "\n" + 
        "7- Reaper" + "\n" + 
        "8- Dragoon" + "\n" + 
        "9- Machinist" + "\n" +
        "10- Bard" + "\n" +  
        "11- Dancer" + "\n" + 
        "12- WhiteMage" + "\n" + 
        "13- Scholar" + "\n" + 
        "14- Astrologian" + "\n" + 
        "15- Sage" + "\n" + 
        "16- DarkKnight" + "\n" + 
        "17- Gunbreaker" + "\n" + 
        "18- Warrior" + "\n" + 
        "19- Paladin" + "\n" + 
        "20- Done" + "\n" + 
        "==================================================================="
        )
        user_int = int(AskInput(20))


        match user_int:
            case 1: player_list += [JobEnum.BlackMage]
            case 2: player_list += [JobEnum.RedMage]
            case 3: player_list += [JobEnum.Summoner]
            case 4: player_list += [JobEnum.Ninja]
            case 5: player_list += [JobEnum.Samurai]
            case 6: player_list += [JobEnum.Monk]
            case 7: player_list += [JobEnum.Reaper]
            case 8: player_list += [JobEnum.Dragoon]
            case 9: player_list += [JobEnum.Machinist]
            case 10: player_list += [JobEnum.Bard]
            case 11: player_list += [JobEnum.Dancer]
            case 12: player_list += [JobEnum.WhiteMage]
            case 13: player_list += [JobEnum.Scholar]
            case 14: player_list += [JobEnum.Astrologian]
            case 15: player_list += [JobEnum.Sage]
            case 16: player_list += [JobEnum.DarkKnight]
            case 17: player_list += [JobEnum.Gunbreaker]
            case 18: player_list += [JobEnum.Warrior]
            case 19: player_list += [JobEnum.Paladin]
            case 20: break

    GenerateLayoutBackend(player_list, namefile)


def MergeFight():
    os.system('CLS') #clearing HUD
    print(
        "===================== MERGING TWO FIGHTS =====================" + "\nWARNING : This function does not make sure every player has a different playerID. So once the merging is done\nmake sure they all have unique playerID." 
    )

    cur_dir = os.getcwd()

    #the saved directory should be in that same folder

    saved_dir = cur_dir + "\\saved"

    saved_fight = os.listdir(saved_dir)
    number_saved_fight = len(saved_fight)
    parent_fight = ""
    child_fight = ""

    if number_saved_fight == 0: #No fight saved
        print("No saved fights were found. Hit any button to return to the main menu.")
        input("...")
        return
    else: #Saved fight
        print(
            "Select one of the saved fights as a parent (will merge into this one) : "+ "\n" + 
            "=========================================================="
            )
        for i in range(1,number_saved_fight+1):
            print(str(i) + " -> " + saved_fight[i-1])

        print("==========================================================")

        userInput = int(AskInput(number_saved_fight))

        parent_fight = saved_fight[userInput-1] #Selecting fight

    
        print(
            "Select one of the saved fights as a child (this one will be merged into) : "+ "\n" + 
            "=========================================================="
            )
        for i in range(1,number_saved_fight+1):
            print(str(i) + " -> " + saved_fight[i-1])

        print("==========================================================")

        userInput = int(AskInput(number_saved_fight))

        child_fight = saved_fight[userInput-1] #Selecting fight

        MergeFightBackEnd(saved_dir + "\\" + child_fight, saved_dir + "\\" + parent_fight, parent_fight)

        


def MainMenu():
    os.system('CLS') #clearing HUD
    #Welcome Message
    print(
    "===================== FFXIV Combat Simulator ======================" + "\n" + 
    "MAIN MENU (input what you want and press ENTER)" + "\n" + 
    "===================================================================" + "\n" + 
    #"1- Simulate fight in code memory" + "\n" + 
    #"2- Save fight in code memory"  + "\n" + 
    "1- Simulate a saved fight" + "\n" + 
    "2- Merge two saved fights" + "\n" + 
    "3- Import fight from FFLogs (Experimental)" + "\n" + 
    "4- Create JSON template file" + "\n" + 
    "5- Credits" + "\n" + 
    "6- Exit" + "\n" + 
    "==================================================================="
    )
    user_input = AskInput(6)

    #if user_input == "1" : ExecuteMemoryCode(False) #Simulating
    #elif user_input == "2": ExecuteMemoryCode(True) #Saving
    if user_input == "1" : SimulateFightMemory()
    elif user_input == "2" : MergeFight() 
    elif user_input == "3" : ImportFight()
    elif user_input == "4" : GenerateLayout()
    elif user_input == "5" : Credits() #Closes program
    elif user_input == "6" : exit() #Closes program
    

def TUI_draw():
    while True : MainMenu() #Draws Main Menu

#This python file will serve as a GUI for the time I do not have an actual GUI
if __name__ == "__main__" : 
    #working_dir = r" INSERT WORKING DIRECTORY HERE "
    #os.chdir(working_dir)
    while True : MainMenu() #Draws Main Menu





from ffxivcalc.Jobs.Base_Spell import Potion, Spell, empty
from ffxivcalc.Jobs.Melee.Melee_Spell import ArmLength
from ffxivcalc.Jobs.Player import MitBuff
from ffxivcalc.Jobs.PlayerEnum import JobEnum
Lock = 0.75



class TankSpell(Spell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)


#########################################
########## WARRIOR Spell  ###############
#########################################

def BeastGaugeRequirement(Player, Spell):
    RemoveBeast(Player, Spell.Cost)
    return Player.BeastGauge >= 0, -1

class WarriorSpell(TankSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, Cost, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

        self.Requirement += [BeastGaugeRequirement] 
        self.Cost = Cost

def RemoveBeast(Player, Gauge):
    Player.BeastGauge -= Gauge #Caanot go under 0 cuz verify if enough gauge


#########################################
########## DARK KNIGHT SKILLS ###########
#########################################

class DRKSkill(TankSpell):
    #A class for Dark Knight Skills containing all the relevant weaponskills/spells, cooldowns,
    #as well as their effects and requirements. For now does not consider out of combo actions.

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, BloodCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

        self.BloodCost = BloodCost
#########################################
########## PALADIN SKILLS  ##############
#########################################

class PaladinSpell(TankSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, isPhysical, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

        self.isPhysical = isPhysical #To know to what ability we will give FoF



#########################################
########## GUNBREAKER SKILLS  ###########
#########################################

def PowderRequirement(Player, Spell):
    Player.PowderGauge -= Spell.PowderCost
    return Player.PowderGauge >=0, 0

class GunbreakerSpell(TankSpell):

    def __init__(self, id, GCD, RecastTime, Potency, Effect, Requirement, PowderCost , type = 0):
        super().__init__(id, GCD, 0, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.PowderCost = PowderCost
        self.Requirement += [PowderRequirement]

#Class Action

#Requirement

def RampartRequirement(Player, Spell):
    return Player.RampartCD <= 0, Player.RampartCD

def LowBlowRequirement(Player, Spell):
    return Player.LowBlowCD <= 0, Player.LowBlowCD

def ProvokeRequirement(Player, Spell):
    return Player.ProvokeCD <= 0, Player.ProvokeCD

def InterjectRequirement(Player, Spell):
    return Player.InterjectCD <= 0, Player.InterjectCD

def ReprisalRequirement(Player, Spell):
    return Player.ReprisalCD <= 0, Player.ReprisalCD
    
def ShirkRequirement(Player, Spell):
    return Player.ShirkCD <= 0, Player.ShirkCD

def BigMitRequirement(Player, Spell):
    return Player.BigMitCD <= 0, Player.BigMitCD

def TankStanceRequirement(Player, Spell):
    return Player.TankStanceCD <= 0, Player.TankStanceCD

#Apply

def ApplyTankStance(Player, Enemy):
    if not Player.TankStanceOn:
        # Only goes in cooldown if the tank stance was already on
        Player.TankStanceOn = True
    else:
        Player.TankStanceCD = 3
        Player.TankStanceOn = False

def ApplyTurnOffTankStance(Player, Enemy):
    Player.TankStanceOn = True

def ApplyBigMit(Player, Enemy):
    Player.BigMitCD = 120
    BigMitBuff = MitBuff(0.7, 15, Player)
    Player.MitBuffList.append(BigMitBuff) # Appends buff

    # Keeps pointer to buff if Warrior
    if Player.JobEnum == JobEnum.Warrior: Player.VengeanceBuff = BigMitBuff
    elif Player.JobEnum == JobEnum.Paladin: Player.BigMitTimer = 15


def ApplyRampart(Player, Enemy):
    Player.RampartCD = 90
    RampartBuff = MitBuff(0.8, 20, Player)
    Player.MitBuffList.append(RampartBuff)
    if Player.JobEnum == JobEnum.Paladin: Player.RampartTimer = 20

def ApplyLowBlow(Player, Enemy):
    Player.LowBlowCD = 25

def ApplyProvoke(Player, Enemy):
    Player.ProvokeCD = 30
    Player.TotalEnemity = Player.CurrentFight.GetEnemityList(1)[0].TotalEnemity + 300
    # Gives enemity to the tank equal to the maximum enemity + 10.
    # The values here are arbitrary. 10 enemity corresponds to 100'000 tank damage (with tank stance on)

def ApplyInterject(Player, Enemy):
    Player.InterjectCD = 30

def ApplyReprisal(Player, Enemy):
    Player.ReprisalCD = 60
    Enemy.Reprisal = True
    Enemy.ReprisalTimer = 10





#ArmLength in Melee_Spell.py
Rampart = TankSpell(7531, False, Lock, 0, 0, 0, ApplyRampart, [RampartRequirement])
LowBlow = TankSpell(7540, False, Lock, 0, 0, 0, ApplyLowBlow, [LowBlowRequirement])
Provoke = TankSpell(7533, False, Lock, 0, 0, 0, ApplyProvoke, [ProvokeRequirement])
Interject = TankSpell(10101010, False, Lock, 0, 0, 0, ApplyInterject, [InterjectRequirement])
Reprisal = TankSpell(7535, False, Lock, 0, 0, 0, ApplyReprisal, [ReprisalRequirement])
RoyalGuard = TankSpell(16142, False, 0, 0, 0, 0, ApplyTankStance, [TankStanceRequirement]) #Turn on Tank Stance
Grit = TankSpell(3629, False, 0, 0, 0, 0, ApplyTankStance, [TankStanceRequirement])
IronWill = TankSpell(28, False, 0, 0, 0, 0, ApplyTankStance, [TankStanceRequirement])
Defiance = TankSpell(48, False, 0, 0, 0, 0, ApplyTankStance, [TankStanceRequirement])
TurnOffTankStance = TankSpell(0, False, 0, 0, 0, 0, ApplyTurnOffTankStance, [])#Turn off Tank Stance

def Shirk(Target):
    """This function is used to generate the shirk action with a customized target.

    Target (Player) : Target of the shirk.
    
    """

    def ApplyShirk(Player, Enemy):
        Player.ShirkCD = 120

        Target.TotalEnemity += Player.TotalEnemity * 0.25
        # Giving 25% of the Player's enemity to the target

        Player.TotalEnemity *= 0.75
        # Loosing 25% of the player's enemity

    custom_shirk = TankSpell(7537, False, Lock, 0, 0, 0, ApplyShirk, [ShirkRequirement])
    custom_shirk.TargetID = Target.playerID
    return custom_shirk


# Limit Break actions

LB1Timer = 1.93
LB2Timer = 3.86
LB3Timer = 3.86

TankLB1 = TankSpell(1111, False,LB1Timer, LB1Timer, 0, 0, empty, [], type=3)
TankLB2 = TankSpell(1112, False,LB2Timer, LB2Timer, 0, 0, empty, [], type=3)
TankLB3 = TankSpell(1113, False,LB3Timer, LB3Timer, 0, 0, empty, [], type=3)

TankAbility = {
10101010 : Interject,
7531 : Rampart,
7535 : Reprisal, 
7548 : ArmLength,
7540 : LowBlow,
7533 : Provoke,
7537 : Shirk,
16142 : RoyalGuard, #Gunbreaker Tank Stance
3629 : Grit, #DarkKnight Tank Stance
48 : Defiance, #Warrior Tank Stance
28 : IronWill, #Paladin Tank Stance
34590541 : Potion,
34592395 : Potion,
34594159 : Potion, 
34594248 : Potion, # Idk what potion this is.
-2 : Potion,
1111 : TankLB1,
1112 : TankLB2,
1113 : TankLB3
}


from ffxivcalc.Jobs.Base_Spell import ManaRequirement, Potion, Spell, empty
from ffxivcalc.Jobs.PlayerEnum import JobEnum
Lock = 0
class CasterSpell(Spell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)


#########################################
########## BLACKMAGE SPELL ##############
#########################################
class BLMSpell(CasterSpell):
    #This class will be all BlackMage Ability
    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, IsFire, IsIce, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)#Calls constructor of Spell

        #BLM specific part

        self.IsFire = IsFire
        self.IsIce = IsIce

#########################################
########## REDMAGE SPELL ################
#########################################


class RedmageSpell(CasterSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, BlackCost, WhiteCost, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

        self.BlackCost = BlackCost
        self.WhiteCost = WhiteCost

#########################################
########## SUMMONER SPELL ###############
#########################################


class SummonerSpell(CasterSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)


#Class Action

#Requirement

def SwiftcastRequirement(Player, Spell):
    return Player.SwiftcastCD <= 0, Player.SwiftcastCD

def LucidDreamingRequirement(Player, Spell):
    #if Player.LucidDreamingCD > 0: input("LUCID IS OUT OF SYNC BY : " + str(Player.LucidDreamingCD))
    #return Player.LucidDreamingCD <= 0, Player.LucidDreamingCD
    return True, -1

def SurecastRequirement(Player, Spell):
    return Player.SurecastCD <= 0, Player.SurecastCD

def AddleRequirement(Player, Spell):
    return Player.AddleCD <= 0, Player.AddleCD

#Apply

def ApplyLucidDreaming(Player, Enemy):
    Player.LucidDreamingCD = 60
    Player.LucidDreamingTimer = 21
    Player.EffectCDList.append(LucidDreamingCheck)


    # The order of insta cast usage for blackmage is SwiftCast -> TripleCast
    # The order for redmage is Acceleration -> Dualcast -> Swiftcast.
    # For purpose of simplicity, the order of these for RDM might not always be like that in sim. For example,
    # if execute Jolt -> Acc -> Swift : The usage will be Acc -> Swift -> Dual.
    # That is because Acc is insert(0), Swift is insert(1) and dual is index 0 at the start.
    # However if execute Jolt -> Swift -> Acc, then the correct use of Acc -> Dual -> Swift will happen.

def ApplySwiftcast(Player, Enemy):
    Player.SwiftcastCD = 60
                             # Inserting in index 1 in order to have it go after ElementalEffect for blm and dualcast effect for rdm.
    Player.EffectList.insert(1,SwiftcastEffect)

def ApplySurecast(Player, Enemy):
    Player.SurecastCD = 120

def ApplyAddle(Player, Enemy):
    Player.AddleCD = 90
    Enemy.Addle = True
    Enemy.AddleTimer = 10


#Effect

def SwiftcastEffect(Player, Spell):
    if Spell.GCD and Spell.CastTime > Lock and not (Player.JobEnum == JobEnum.BlackMage and Spell.id == 152 and Player.Fire3Proc):  #If GCD and not already insta cast and not F3 Proc when casting F3
        Spell.CastTime = Lock
        Player.EffectToRemove.append(SwiftcastEffect)

#Check

def LucidDreamingCheck(Player, Enemy):
    #input("got in")
    if (int(Player.LucidDreamingTimer * 100)/100)%3 == 0 and Player.LucidDreamingTimer != 21:
        #if on a tic and not equal to 21
        #input("got in")

        if Player.JobEnum == JobEnum.BlackMage:
            #We have to check if in firephase, in which case no mana regen
            if Player.ElementalGauge <= 0: #If in ice phase
                #input("Adding mana")
                Player.Mana = min(10000, Player.Mana + 550)
        else:
            #Then any other player
            Player.Mana = min(10000, Player.Mana + 550)

        #Check if we are done
        if Player.LucidDreamingTimer <= 0:
            Player.EffectToRemove.append(LucidDreamingCheck)

#Class Action
Swiftcast = CasterSpell(7561, False,0, Lock, 0, 0, ApplySwiftcast, [SwiftcastRequirement])
LucidDreaming = CasterSpell(7562, False, Lock,0,0, 0, ApplyLucidDreaming, [LucidDreamingRequirement])
Surecast = CasterSpell(7559, False, Lock, 0, 0, 0, ApplySurecast, [SurecastRequirement])
Addle = CasterSpell(7560, False, Lock, 0, 0, 0, ApplyAddle, [AddleRequirement])
Sleep = CasterSpell(25880, True, 2.5, 2.5, 0, 800, empty, [ManaRequirement])

# Limit Break actions

LB1Timer = 5.1
LB2Timer = 8.1
LB3Timer = 12.6

CasterLB1 = CasterSpell(1111, False,LB1Timer, LB1Timer, 0, 0, empty, [], type=3)
CasterLB2 = CasterSpell(1112, False,LB2Timer, LB2Timer, 0, 0, empty, [], type=3)
CasterLB3 = CasterSpell(1113, False,LB3Timer, LB3Timer, 0, 0, empty, [], type=3)

CasterAbility = {
7561 : Swiftcast, 
7562 : LucidDreaming,  
7559 : Surecast, 
7560 : Addle, 
25880 : Sleep,
34590544 : Potion, #Grade 6
34592398 : Potion, # Grade 7
34594162 : Potion, # Grade 8
-2 : Potion,
1111 : CasterLB1,
1112 : CasterLB2,
1113 : CasterLB3
}